import json
import os
from dataclasses import dataclass
from unittest.mock import patch

import appdirs
import pytest
from git import Head

from flexlate import branch_update
from flexlate.add_mode import AddMode
from flexlate.config import FlexlateConfig, FlexlateProjectConfig
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.exc import TemplateSourceWithNameAlreadyExistsException
from flexlate.ext_git import repo_has_merge_conflicts
from flexlate.template.base import Template
from flexlate.template.copier import CopierTemplate
from flexlate.template.types import TemplateType
from flexlate.transactions.transaction import FlexlateTransaction
from tests.config import GENERATED_FILES_DIR
from tests.fileutils import cookiecutter_one_generated_text_content
from tests.fixtures.git import *
from tests.fixtures.subdir_style import SubdirStyle, subdir_style
from tests.fixtures.template import *
from tests.fixtures.templated_repo import *
from tests.fixtures.add_mode import add_mode
from tests.fixtures.transaction import *
from tests.fs_checks import (
    assert_template_source_cookiecutter_one_added_correctly,
    assert_cookiecutter_one_applied_template_added_correctly,
)
from tests.gitutils import (
    accept_theirs_in_merge_conflict,
    assert_main_commit_message_matches,
)


def test_add_template_source_to_repo(
    repo_with_placeholder_committed: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_placeholder_committed
    adder = Adder()
    adder.add_template_source(
        repo,
        cookiecutter_one_template,
        add_source_transaction,
        out_root=GENERATED_REPO_DIR,
        target_version="some version",
    )
    assert_template_source_cookiecutter_one_added_correctly(
        cookiecutter_one_template, target_version="some version"
    )


def test_add_template_source_with_existing_name_fails(
    repo_with_cookiecutter_one_template_source: Repo,
    copier_one_template: CopierTemplate,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    adder = Adder()
    with pytest.raises(TemplateSourceWithNameAlreadyExistsException):
        adder.add_template_source(
            repo,
            copier_one_template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_add_local_cookiecutter_applied_template_to_repo(
    add_mode: AddMode,
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    template = cookiecutter_one_template
    adder = Adder()
    adder.apply_template_and_add(
        repo,
        template,
        add_output_transaction,
        out_root=GENERATED_REPO_DIR,
        add_mode=add_mode,
        no_input=True,
    )

    if add_mode == AddMode.USER:
        config_dir = GENERATED_FILES_DIR
        template_root = GENERATED_REPO_DIR.absolute()
    elif add_mode == AddMode.PROJECT:
        config_dir = GENERATED_REPO_DIR
        template_root = Path(".")
    elif add_mode == AddMode.LOCAL:
        # Template has output in a subdir, so with
        # local mode config will also be in the subdir
        config_dir = GENERATED_REPO_DIR / "b"
        template_root = Path("..")
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    assert_cookiecutter_one_applied_template_added_correctly(
        template, config_dir, template_root, add_mode
    )


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_add_local_copier_output_subdir_applied_template_to_repo(
    add_mode: AddMode,
    repo_with_copier_output_subdir_template_source: Repo,
    copier_output_subdir_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_copier_output_subdir_template_source
    template = copier_output_subdir_template
    adder = Adder()
    adder.apply_template_and_add(
        repo,
        template,
        add_output_transaction,
        out_root=GENERATED_REPO_DIR,
        add_mode=add_mode,
        no_input=True,
    )

    if add_mode == AddMode.USER:
        config_dir = GENERATED_FILES_DIR
        template_root = GENERATED_REPO_DIR.absolute()
    elif add_mode == AddMode.PROJECT:
        config_dir = GENERATED_REPO_DIR
        template_root = Path(".")
    elif add_mode == AddMode.LOCAL:
        # Even though template has output in a subdir, with copier
        # it all still renders at root in output
        config_dir = GENERATED_REPO_DIR
        template_root = Path(".")
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    config_path = config_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"qone": "aone", "qtwo": "atwo"}
    assert at.root == template_root
    assert at.add_mode == add_mode


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_add_remote_cookiecutter_applied_template_to_repo(
    add_mode: AddMode,
    repo_with_remote_cookiecutter_template_source: Repo,
    cookiecutter_remote_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_remote_cookiecutter_template_source
    template = cookiecutter_remote_template
    adder = Adder()
    adder.apply_template_and_add(
        repo,
        template,
        add_output_transaction,
        out_root=GENERATED_REPO_DIR,
        add_mode=add_mode,
        no_input=True,
    )

    if add_mode == AddMode.USER:
        config_dir = GENERATED_FILES_DIR
        template_root = GENERATED_REPO_DIR.absolute()
    elif add_mode == AddMode.PROJECT:
        config_dir = GENERATED_REPO_DIR
        template_root = Path(".")
    elif add_mode == AddMode.LOCAL:
        config_dir = GENERATED_REPO_DIR / "abc"
        template_root = Path("..")
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    _assert_remote_cookiecutter_applied_correctly(
        template, template_root, add_mode, config_dir=config_dir
    )

    template_sources_config_path = GENERATED_REPO_DIR / "flexlate.json"
    ts_config = FlexlateConfig.load(template_sources_config_path)
    assert len(ts_config.template_sources) == 1
    source = ts_config.template_sources[0]
    assert source.name == cookiecutter_remote_template.name
    assert source.path == cookiecutter_remote_template.git_url
    assert source.version == cookiecutter_remote_template.version
    assert source.type == TemplateType.COOKIECUTTER
    assert source.render_relative_root_in_output == Path("{{ cookiecutter.name }}")
    assert source.render_relative_root_in_template == Path("{{ cookiecutter.name }}")


def _assert_remote_cookiecutter_applied_correctly(
    template: Template,
    template_root: Path,
    add_mode: AddMode,
    config_dir: Path = GENERATED_REPO_DIR,
):
    config_path = config_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"name": "abc", "key": "value"}
    assert at.root == template_root
    assert at.add_mode == add_mode


def test_add_source_and_output_at_target_version(
    repo_with_placeholder_committed: Repo,
    cookiecutter_remote_version_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_placeholder_committed
    template = cookiecutter_remote_version_one_template

    adder = Adder()
    adder.add_template_source(
        repo,
        template,
        add_source_transaction,
        out_root=GENERATED_REPO_DIR,
        target_version=COOKIECUTTER_REMOTE_VERSION_1,
    )
    adder.apply_template_and_add(
        repo,
        template,
        add_output_transaction,
        out_root=GENERATED_REPO_DIR,
        no_input=True,
        add_mode=AddMode.PROJECT,
    )

    # Check for version 1 content
    output_path = GENERATED_REPO_DIR / "abc" / "abc.txt"
    assert output_path.read_text() == "value"

    # Check for version 1 in configs
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    ts = config.template_sources[0]
    assert ts.version == COOKIECUTTER_REMOTE_VERSION_1
    assert ts.target_version == COOKIECUTTER_REMOTE_VERSION_1
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.version == COOKIECUTTER_REMOTE_VERSION_1


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_add_applied_template_to_subdir(
    add_mode: AddMode,
    subdir_style: SubdirStyle,
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    template = cookiecutter_one_template
    subdir = GENERATED_REPO_DIR / "subdir1" / "subdir2"
    subdir.mkdir(parents=True)
    adder = Adder()
    if subdir_style == SubdirStyle.CD:
        with change_directory_to(subdir):
            adder.apply_template_and_add(
                repo, template, add_output_transaction, add_mode=add_mode, no_input=True
            )
    elif subdir_style == SubdirStyle.PROVIDE_RELATIVE:
        with change_directory_to(GENERATED_REPO_DIR):
            adder.apply_template_and_add(
                repo,
                template,
                add_output_transaction,
                out_root=subdir.relative_to(os.getcwd()),
                add_mode=add_mode,
                no_input=True,
            )
    elif subdir_style == SubdirStyle.PROVIDE_ABSOLUTE:
        adder.apply_template_and_add(
            repo,
            template,
            add_output_transaction,
            out_root=subdir.absolute(),
            add_mode=add_mode,
            no_input=True,
        )

    if add_mode == AddMode.LOCAL:
        config_dir = subdir / "b"
        template_root = Path("..")
    elif add_mode == AddMode.PROJECT:
        config_dir = GENERATED_REPO_DIR
        template_root = subdir.relative_to(GENERATED_REPO_DIR)
    elif add_mode == AddMode.USER:
        config_dir = GENERATED_FILES_DIR
        template_root = subdir.absolute()
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    config_path = config_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"a": "b", "c": ""}
    assert at.root == template_root
    assert at.add_mode == add_mode

    output_file_path = subdir / "b" / "text.txt"
    assert output_file_path.read_text() == "b"


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_add_multiple_applied_templates_for_one_source(
    add_mode: AddMode,
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    template = cookiecutter_one_template
    subdir = GENERATED_REPO_DIR / "subdir1" / "subdir2"
    subdir.mkdir(parents=True)
    adder = Adder()
    with change_directory_to(GENERATED_REPO_DIR):
        adder.apply_template_and_add(
            repo, template, add_output_transaction, add_mode=add_mode, no_input=True
        )
    with change_directory_to(subdir):
        adder.apply_template_and_add(
            repo, template, add_output_transaction, add_mode=add_mode, no_input=True
        )

    @dataclass
    class OutputOptions:
        config_dir: Path
        template_root: Path
        render_root: Path
        expect_num_applied_templates: int = 1
        applied_template_index: int = 0

    output_options: List[OutputOptions] = []
    if add_mode == AddMode.LOCAL:
        output_options.extend(
            [
                OutputOptions(GENERATED_REPO_DIR / "b", Path(".."), GENERATED_REPO_DIR),
                OutputOptions(subdir / "b", Path(".."), subdir),
            ]
        )
    elif add_mode == AddMode.PROJECT:
        output_options.extend(
            [
                OutputOptions(
                    GENERATED_REPO_DIR,
                    Path("."),
                    GENERATED_REPO_DIR,
                    expect_num_applied_templates=2,
                ),
                OutputOptions(
                    GENERATED_REPO_DIR,
                    subdir.relative_to(GENERATED_REPO_DIR),
                    subdir,
                    expect_num_applied_templates=2,
                    applied_template_index=1,
                ),
            ]
        )
    elif add_mode == AddMode.USER:
        output_options.extend(
            [
                OutputOptions(
                    GENERATED_FILES_DIR,
                    GENERATED_REPO_DIR.absolute(),
                    GENERATED_REPO_DIR,
                    expect_num_applied_templates=2,
                ),
                OutputOptions(
                    GENERATED_FILES_DIR,
                    subdir.absolute(),
                    subdir,
                    expect_num_applied_templates=2,
                    applied_template_index=1,
                ),
            ]
        )
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    for output_option in output_options:
        config_dir = output_option.config_dir
        template_root = output_option.template_root
        render_root = output_option.render_root
        expect_num_applied_templates = output_option.expect_num_applied_templates
        applied_template_index = output_option.applied_template_index
        config_path = config_dir / "flexlate.json"
        config = FlexlateConfig.load(config_path)
        assert len(config.applied_templates) == expect_num_applied_templates
        at = config.applied_templates[applied_template_index]
        assert at.name == template.name
        assert at.version == template.version
        assert at.data == {"a": "b", "c": ""}
        assert at.root == template_root
        assert at.add_mode == add_mode

        output_file_path = render_root / "b" / "text.txt"
        assert output_file_path.read_text() == "b"


def test_add_source_to_project_with_existing_outputs(
    repo_with_cookiecutter_one_template_source_and_output: Repo,
    cookiecutter_two_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source_and_output
    adder = Adder()
    adder.add_template_source(
        repo,
        cookiecutter_two_template,
        add_source_transaction,
        out_root=GENERATED_REPO_DIR,
        target_version="some version",
    )
    assert cookiecutter_one_generated_text_content(gen_dir=GENERATED_REPO_DIR) == "b"
    source_config_path = GENERATED_REPO_DIR / "flexlate.json"
    source_config = FlexlateConfig.load(source_config_path)
    assert len(source_config.applied_templates) == 0
    assert len(source_config.template_sources) == 2
    source = source_config.template_sources[1]
    assert source.name == cookiecutter_two_template.name
    assert source.path == str(cookiecutter_two_template.path)
    assert source.version == cookiecutter_two_template.version
    assert source.type == TemplateType.COOKIECUTTER
    assert source.target_version == "some version"
    assert source.render_relative_root_in_output == Path("{{ cookiecutter.a }}")
    assert source.render_relative_root_in_template == Path("{{ cookiecutter.a }}")
    at_config_path = GENERATED_REPO_DIR / "b" / "flexlate.json"
    at_config = FlexlateConfig.load(at_config_path)
    assert len(at_config.applied_templates) == 1


def test_add_source_with_merge_conflicts_and_resolution(
    repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation,
    cookiecutter_one_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation

    adder = Adder()

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        assert repo_has_merge_conflicts(repo)
        accept_theirs_in_merge_conflict(repo)
        stage_and_commit_all(repo, "Manually resolve conflicts")
        return True

    with patch.object(branch_update, "confirm_user", _resolve_conflicts_then_type_yes):
        adder.add_template_source(
            repo,
            cookiecutter_one_template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
            target_version="some version",
        )

    assert_template_source_cookiecutter_one_added_correctly(
        cookiecutter_one_template,
        num_sources=2,
        source_idx=1,
        num_applied_templates=1,
        target_version="some version",
    )


def test_add_source_with_merge_conflicts_and_reject(
    repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation,
    cookiecutter_one_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation

    adder = Adder()

    def _reject(prompt: str) -> bool:
        return False

    with patch.object(branch_update, "confirm_user", _reject):
        adder.add_template_source(
            repo,
            cookiecutter_one_template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
            target_version="some version",
        )

    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1

    assert_main_commit_message_matches(
        repo.commit().message, "Reformat flexlate config"
    )
    for branch_name in [DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME]:
        branch = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        assert_main_commit_message_matches(
            repo.commit().message, "Update flexlate templates"
        )


def test_add_output_with_merge_conflicts_and_resolution(
    repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation: Repo,
    cookiecutter_remote_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation
    template = cookiecutter_remote_template

    adder = Adder()

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        assert repo_has_merge_conflicts(repo)
        accept_theirs_in_merge_conflict(repo)
        stage_and_commit_all(repo, "Manually resolve conflicts")
        return True

    subdir = GENERATED_REPO_DIR / "subdir"
    subdir.mkdir()

    with patch.object(branch_update, "confirm_user", _resolve_conflicts_then_type_yes):
        with change_directory_to(subdir):
            adder.apply_template_and_add(
                repo,
                template,
                add_output_transaction,
                no_input=True,
            )

    _assert_remote_cookiecutter_applied_correctly(
        template, Path(".."), AddMode.LOCAL, config_dir=subdir / "abc"
    )


def test_add_project_config_with_git(repo_with_placeholder_committed: Repo):
    repo = repo_with_placeholder_committed
    adder = Adder()
    adder.init_project_and_add_to_branches(repo)

    for branch_name in [
        "master",
        DEFAULT_MERGED_BRANCH_NAME,
        DEFAULT_TEMPLATE_BRANCH_NAME,
    ]:
        branch: Head = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        config = FlexlateProjectConfig.load(
            GENERATED_REPO_DIR / "flexlate-project.json"
        )
        assert len(config.projects) == 1
        project = config.projects[0]
        assert project.path == Path(".")
        assert project.default_add_mode == AddMode.LOCAL
        assert project.merged_branch_name == DEFAULT_MERGED_BRANCH_NAME
        assert project.template_branch_name == DEFAULT_TEMPLATE_BRANCH_NAME
        assert project.remote == "origin"


def test_init_project_from_template_source_path_remote_cookiecutter(
    cookiecutter_remote_template: CookiecutterTemplate,
    add_source_and_output_transaction: FlexlateTransaction,
):
    template = cookiecutter_remote_template
    transaction = add_source_and_output_transaction

    adder = Adder()
    with change_directory_to(GENERATED_FILES_DIR):
        folder_name = adder.init_project_from_template_source_path(
            template, transaction, no_input=True
        )

    project_dir = GENERATED_FILES_DIR / folder_name
    # Ensure project is a git repo
    repo = Repo(project_dir)
    assert (
        repo.commit().message
        == "Merge branch 'flexlate-templates' into flexlate-output\n"
    )

    content_path = project_dir / "abc.txt"
    content = content_path.read_text()
    assert content == "some new header\nvalue"

    non_content_path = project_dir / "README.md"
    assert not non_content_path.exists()

    config_path = project_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    source = config.template_sources[0]
    assert source.name == template.name
    assert source.path == template.git_url
    assert source.version == template.version
    assert source.type == TemplateType.COOKIECUTTER
    assert source.render_relative_root_in_output == Path("{{ cookiecutter.name }}")
    assert source.render_relative_root_in_template == Path("{{ cookiecutter.name }}")
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"name": "abc", "key": "value"}
    assert at.root == Path("..")
    assert at.add_mode == AddMode.LOCAL


def test_init_project_from_template_source_path_local_copier(
    copier_one_template: CopierTemplate,
    add_source_and_output_transaction: FlexlateTransaction,
):
    template = copier_one_template
    transaction = add_source_and_output_transaction

    adder = Adder()
    with change_directory_to(GENERATED_FILES_DIR):
        folder_name = adder.init_project_from_template_source_path(
            template, transaction, no_input=True
        )

    project_dir = GENERATED_FILES_DIR / folder_name

    _assert_init_from_local_copier_output_correct(project_dir, template)


def test_init_project_from_template_source_path_local_copier_in_another_directory(
    copier_one_template: CopierTemplate,
    add_source_and_output_transaction: FlexlateTransaction,
):
    template = copier_one_template
    transaction = add_source_and_output_transaction

    adder = Adder()
    with change_directory_to(GENERATED_FILES_DIR):
        subdir = GENERATED_FILES_DIR / "subdir"
        folder_name = adder.init_project_from_template_source_path(
            template, transaction, path=subdir, no_input=True
        )

    project_dir = subdir / folder_name

    _assert_init_from_local_copier_output_correct(project_dir, template)


def _assert_init_from_local_copier_output_correct(
    project_dir: Path, template: Template
):
    # Ensure project is a git repo
    repo = Repo(project_dir)
    assert_main_commit_message_matches(
        repo.commit().message, "Update flexlate templates"
    )

    content_path = project_dir / "a1.txt"
    content = content_path.read_text()
    assert content == "1"

    readme_path = project_dir / "README.md"
    assert readme_path.read_text() == "some existing content"

    config_path = project_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    source = config.template_sources[0]
    assert source.name == template.name
    assert source.path == str(template.path)
    assert source.version == template.version
    assert source.type == TemplateType.COPIER
    assert source.render_relative_root_in_output == Path(".")
    assert source.render_relative_root_in_template == Path(".")
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"q1": "a1", "q2": 1, "q3": None}
    assert at.root == Path(".")
    assert at.add_mode == AddMode.LOCAL



from pathlib import Path
from typing import Optional

from flexlate.template.base import Template
from flexlate.template.types import TemplateType
from flexlate.template_config.copier import CopierConfig


class CopierTemplate(Template):
    _type = TemplateType.COPIER

    def __init__(
        self,
        config: CopierConfig,
        path: Path,
        name: Optional[str] = None,
        version: Optional[str] = None,
        target_version: Optional[str] = None,
        git_url: Optional[str] = None,
        template_source_path: Optional[str] = None,
        render_relative_root_in_template: Path = Path("."),
    ):
        super().__init__(
            config,
            path,
            name=name,
            version=version,
            target_version=target_version,
            git_url=git_url,
            template_source_path=template_source_path,
            render_relative_root_in_template=render_relative_root_in_template,
            # Render relative root in output is always . for Copier
            render_relative_root_in_output=Path("."),
        )



import sys
from contextlib import contextmanager
from functools import wraps
from types import TracebackType
from typing import Callable, Type, Tuple

from flexlate.styles import print_styled, ALERT_STYLE


def simple_output_for_exceptions(*exceptions: Type[BaseException]):
    exception_handler = _create_exception_handler(exceptions)

    def _simple_output_for_exceptions(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with _handle_exceptions_with(exception_handler):
                return func(*args, **kwargs)

        return wrapper

    return _simple_output_for_exceptions


ExceptionHandler = Callable[[Type[BaseException], BaseException, TracebackType], None]


@contextmanager
def _handle_exceptions_with(exc_handler: ExceptionHandler):
    "Sets a custom exception handler for the scope of a 'with' block."
    sys.excepthook = exc_handler  #  type: ignore
    yield
    sys.excepthook = sys.__excepthook__  # type: ignore


def _create_exception_handler(
    exceptions: Tuple[Type[BaseException], ...]
) -> ExceptionHandler:
    def handle_specific_exceptions(
        type_: Type[BaseException], value: BaseException, traceback: TracebackType
    ):
        if isinstance(value, exceptions):
            print_styled(f"{type_.__name__}: {value}", ALERT_STYLE)
        else:
            sys.__excepthook__(type_, value, traceback)

    return handle_specific_exceptions



import pytest
from git import Repo

from flexlate.ext_git import stage_and_commit_all
from tests.config import GENERATED_REPO_DIR
from tests.dirutils import wipe_generated_folder
from flexlate.path_ops import change_directory_to
from tests.gitutils import (
    create_empty_repo,
    add_dummy_file_to_repo,
    add_dummy_file2_to_repo,
    add_gitignore_and_ignored_file_to_repo,
)


@pytest.fixture
def empty_generated_repo() -> Repo:
    wipe_generated_folder()
    yield create_empty_repo()


@pytest.fixture
def repo_with_placeholder_committed(empty_generated_repo: Repo) -> Repo:
    repo = empty_generated_repo
    with change_directory_to(GENERATED_REPO_DIR):
        add_dummy_file_to_repo(repo)
        stage_and_commit_all(repo, "Initial commit")
    yield repo


@pytest.fixture
def dirty_repo(repo_with_placeholder_committed: Repo) -> Repo:
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        add_dummy_file2_to_repo(repo)
    yield repo


@pytest.fixture
def repo_with_gitignore(repo_with_placeholder_committed: Repo) -> Repo:
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        add_gitignore_and_ignored_file_to_repo(repo)
        stage_and_commit_all(repo, "Add sample .gitignore with ignored file")
    yield repo



from typing import Final

from rich.console import Console

console: Final[Console] = Console()


ACTION_REQUIRED_STYLE: Final[str] = "[yellow]:pencil:"
QUESTION_STYLE: Final[str] = ":question:"
ALERT_STYLE: Final[str] = "[red]:x:"
INFO_STYLE: Final[str] = ""
SUCCESS_STYLE: Final[str] = "[green]:heavy_check_mark:"


def styled(message: str, style: str) -> str:
    return f"{style} {message}"


def print_styled(message: str, style: str):
    console.print(styled(message, style))



from typing import Dict, Any, Sequence, List

TemplateData = Dict[str, Any]


def merge_data(
    overrides: Sequence[TemplateData], defaults: Sequence[TemplateData]
) -> List[TemplateData]:
    out_data: List[TemplateData] = []
    for i, default_data in enumerate(defaults):
        try:
            override_data = overrides[i]
        except IndexError:
            override_data = {}
        out_data.append({**default_data, **override_data})
    return out_data



from pathlib import Path

PROJECT_DIR = Path(__file__).parent.parent

TESTS_DIR = Path(__file__).parent
INPUT_FILES_DIR = TESTS_DIR / "input_files"
GENERATED_FILES_DIR = TESTS_DIR / "generated"
GENERATED_REPO_DIR = GENERATED_FILES_DIR / "project"

TEMPLATES_DIR = INPUT_FILES_DIR / "templates"

COOKIECUTTERS_DIR = TEMPLATES_DIR / "cookiecutters"
COOKIECUTTER_ONE_NAME = "one"
COOKIECUTTER_ONE_DIR = COOKIECUTTERS_DIR / COOKIECUTTER_ONE_NAME
COOKIECUTTER_ONE_VERSION = "1c154af24ff30bc4cab8cf9d543304d9"
COOKIECUTTER_ONE_MODIFIED_VERSION = "2dc435b3d7e256fbdcc78e62faaabff4"
COOKIECUTTER_TWO_NAME = "two"
COOKIECUTTER_TWO_DIR = COOKIECUTTERS_DIR / COOKIECUTTER_TWO_NAME
COOKIECUTTER_WITH_HOOKS_NAME = "with-hooks"
COOKIECUTTER_WITH_HOOKS_DIR = COOKIECUTTERS_DIR / COOKIECUTTER_WITH_HOOKS_NAME
COOKIECUTTER_WITH_HOOKS_VERSION = "f54b5faa0a6318d5a21434235eb9e1a9"
COOKIECUTTER_WITH_HOOKS_MODIFIED_VERSION = "534cfcb2dce2a8d94e15e872e6ff08ba"
COOKIECUTTER_REMOTE_URL = (
    "https://github.com/nickderobertis/cookiecutter-simple-example"
)
COOKIECUTTER_REMOTE_NAME = "cookiecutter-simple-example"
COOKIECUTTER_REMOTE_VERSION_1 = "c390901c4fd599473bdb95fa4dd3d2a6eb2b34f0"
COOKIECUTTER_REMOTE_VERSION_2 = "ee8da996b5d74dfcbba5727d7c950c88173dc9ca"

COPIERS_DIR = TEMPLATES_DIR / "copiers"
COPIER_ONE_NAME = "one"
COPIER_ONE_DIR = COPIERS_DIR / COPIER_ONE_NAME
COPIER_ONE_VERSION = "a6e386b97e1d7de2670e4fc4fee5b655"
COPIER_ONE_MODIFIED_VERSION = "c5d65a8f94813d33ef031b597358d085"
COPIER_OUTPUT_SUBDIR_NAME = "output-subdir"
COPIER_OUTPUT_SUBDIR_DIR = COPIERS_DIR / COPIER_OUTPUT_SUBDIR_NAME
COPIER_OUTPUT_SUBDIR_VERSION = "783e158aea9852664d497b19f6ba7b8b"
COPIER_OUTPUT_SUBDIR_MODIFIED_VERSION = "does not matter yet for tests"
COPIER_FROM_COOKIECUTTER_ONE_NAME = "from-cookiecutter-one"
COPIER_FROM_COOKIECUTTER_ONE_DIR = COPIERS_DIR / COPIER_FROM_COOKIECUTTER_ONE_NAME
COPIER_FROM_COOKIECUTTER_ONE_VERSION = "5a4e3d4bf1fe026cb6e63dbca154825a"
COPIER_WITH_TASKS_NAME = "with-tasks"
COPIER_WITH_TASKS_DIR = COPIERS_DIR / COPIER_WITH_TASKS_NAME
COPIER_WITH_TASKS_VERSION = "ba749e5c12a24679aa138dfaec88f8ee"
COPIER_WITH_TASKS_MODIFIED_VERSION = "8597833b6069a1ecb114e5d84cee0d09"
COPIER_REMOTE_URL = "https://github.com/nickderobertis/copier-simple-example"
COPIER_REMOTE_NAME = "copier-simple-example"
COPIER_REMOTE_VERSION_1 = "c7e1ba1bfb141e9c577e7c21ee4a5d3ae5dde04d"
COPIER_REMOTE_VERSION_2 = "f3fa2c9526c12b3011e0a26108ce141373539bca"

CONFIGS_DIR = INPUT_FILES_DIR / "configs"
CONFIG_1_PATH = CONFIGS_DIR / "flexlate.json"
CONFIG_SUBDIR_1 = CONFIG_1_PATH / "subdir1"
CONFIG_SUBDIR_2 = CONFIGS_DIR / "subdir2"

PROJECTS_DIR = INPUT_FILES_DIR / "projects"
NESTED_PROJECT_DIR = PROJECTS_DIR / "nested"

PROJECT_CONFIGS_DIR = INPUT_FILES_DIR / "project_configs"
PROJECT_CONFIGS_ROOT_DIR = PROJECT_CONFIGS_DIR / "root"
PROJECT_CONFIGS_ROOT_CONFIG_PATH = PROJECT_CONFIGS_ROOT_DIR / "flexlate-project.json"
PROJECT_CONFIGS_PROJECT_1_PATH = PROJECT_CONFIGS_ROOT_DIR / "1"
PROJECT_CONFIGS_PROJECT_2_PATH = PROJECT_CONFIGS_ROOT_DIR / "2"
PROJECT_CONFIGS_PROJECT_1_SUBDIR = PROJECT_CONFIGS_PROJECT_1_PATH / "subdir"
PROJECT_CONFIGS_PROJECT_2_SUBDIR = PROJECT_CONFIGS_PROJECT_2_PATH / "subdir"
PROJECT_CONFIGS_1_CONFIG_PATH = PROJECT_CONFIGS_PROJECT_1_PATH / "flexlate-project.json"


if not GENERATED_FILES_DIR.exists():
    GENERATED_FILES_DIR.mkdir()



from typing import List, Final

import pytest

from flexlate.add_mode import AddMode

all_add_modes: Final[List[AddMode]] = list(AddMode)


@pytest.fixture(scope="module", params=all_add_modes)
def add_mode(request) -> AddMode:
    return request.param



from uuid import UUID

import pytest

from flexlate.transactions.transaction import FlexlateTransaction, TransactionType

ADD_SOURCE_ID = UUID("93f984ca-6e8f-45e9-b9b0-aebebfe798c1")
ADD_OUTPUT_ID = UUID("86465f4d-9752-4ae5-aaa7-791b4c814e8d")
ADD_SOURCE_AND_OUTPUT_ID = UUID("bf4cd42c-10b1-4bf9-a15f-294f5be738b0")
REMOVE_SOURCE_ID = UUID("c034ec63-d2b5-4d8c-aef1-f96e29a6f5d1")
REMOVE_OUTPUT_ID = UUID("79715a11-a3c4-40b1-a49b-9d8388e5c28d")
UPDATE_TRANSACTION_ID = UUID("347711b7-3bf9-484e-be52-df488f3cf598")
SYNC_TRANSACTION_ID = UUID("4825ce35-1a03-43de-ad8a-1ecc0ed68b62")
BOOTSTRAP_TRANSACTION_ID = UUID("37c61224-2b8d-4ee5-8846-49d5474a40bd")
UPDATE_TARGET_VERSION_ID = UUID("a5632854-48b4-4f82-904b-bff81dc40b02")


@pytest.fixture
def add_source_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(type=TransactionType.ADD_SOURCE, id=ADD_SOURCE_ID)


@pytest.fixture
def add_output_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(type=TransactionType.ADD_OUTPUT, id=ADD_OUTPUT_ID)


@pytest.fixture
def add_source_and_output_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(
        type=TransactionType.ADD_SOURCE_AND_OUTPUT, id=ADD_SOURCE_AND_OUTPUT_ID
    )


@pytest.fixture
def remove_source_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(type=TransactionType.REMOVE_SOURCE, id=REMOVE_SOURCE_ID)


@pytest.fixture
def remove_output_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(type=TransactionType.REMOVE_OUTPUT, id=REMOVE_OUTPUT_ID)


@pytest.fixture
def update_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(type=TransactionType.UPDATE, id=UPDATE_TRANSACTION_ID)


@pytest.fixture
def sync_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(type=TransactionType.SYNC, id=SYNC_TRANSACTION_ID)


@pytest.fixture
def bootstrap_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(
        type=TransactionType.BOOTSTRAP, id=BOOTSTRAP_TRANSACTION_ID
    )


@pytest.fixture
def update_target_version_transaction() -> FlexlateTransaction:
    yield FlexlateTransaction(
        type=TransactionType.UPDATE_TARGET_VERSION, id=UPDATE_TARGET_VERSION_ID
    )



from enum import Enum
from typing import List, Final, Optional

import pytest


class SubdirStyle(str, Enum):
    CD = "cd"
    PROVIDE_RELATIVE = "provide relative"
    PROVIDE_ABSOLUTE = "provide absolute"


all_subdir_styles: Final[List[SubdirStyle]] = list(SubdirStyle)
all_subdir_styles_or_none: Final[List[Optional[SubdirStyle]]] = [
    None,
    *all_subdir_styles,
]


@pytest.fixture(scope="module", params=all_subdir_styles)
def subdir_style(request) -> SubdirStyle:
    return request.param


@pytest.fixture(scope="module", params=all_subdir_styles_or_none)
def subdir_style_or_none(request) -> Optional[SubdirStyle]:
    return request.param



from functools import partial

from tests.integration.cli_stub import fxt as _fxt, ExceptionHandling

# Ignore exceptions so that we can test error codes instead
fxt = partial(_fxt, exception_handling=ExceptionHandling.IGNORE)



import tempfile
from pathlib import Path

import pytest

from flexlate.finder.multi import MultiFinder
from flexlate.finder.specific.cookiecutter import CookiecutterFinder
from flexlate.finder.specific.copier import CopierFinder
from flexlate.path_ops import change_directory_to
from flexlate.template.base import Template
from flexlate.template.cookiecutter import CookiecutterTemplate
from flexlate.template_path import get_local_repo_path_and_name_cloning_if_repo_url
from tests.config import (
    COOKIECUTTER_ONE_DIR,
    COOKIECUTTER_REMOTE_URL,
    COOKIECUTTER_REMOTE_VERSION_1,
    COOKIECUTTER_REMOTE_VERSION_2,
    COOKIECUTTER_ONE_VERSION,
    COPIER_ONE_DIR,
    COPIER_ONE_VERSION,
    COPIER_REMOTE_VERSION_1,
    COPIER_REMOTE_VERSION_2,
    COPIER_REMOTE_URL,
    COPIER_REMOTE_NAME,
    COOKIECUTTER_REMOTE_NAME,
    COPIER_ONE_NAME,
    COOKIECUTTER_ONE_NAME,
    GENERATED_FILES_DIR,
    COPIER_OUTPUT_SUBDIR_VERSION,
    COPIER_OUTPUT_SUBDIR_NAME,
    COPIER_OUTPUT_SUBDIR_DIR,
)


def test_get_cookiecutter_config():
    finder = CookiecutterFinder()
    config = finder.get_config(COOKIECUTTER_ONE_DIR)
    assert config.defaults == {"a": "b", "c": ""}


def test_get_copier_config():
    finder = CopierFinder()
    config = finder.get_config(COPIER_ONE_DIR)
    assert config.defaults == {"q1": "a1", "q2": 1, "q3": None}


def test_get_cookiecutter_local_template():
    finder = CookiecutterFinder()
    template = finder.find(str(COOKIECUTTER_ONE_DIR), COOKIECUTTER_ONE_DIR)
    _assert_template_matches_cookiecutter_one(template)


def _assert_template_matches_cookiecutter_one(
    template: Template, expect_path: Path = COOKIECUTTER_ONE_DIR
):
    assert template.path == expect_path
    assert template.name == COOKIECUTTER_ONE_NAME
    assert template.git_url is None
    assert template.version == COOKIECUTTER_ONE_VERSION
    assert template.config.defaults == {"a": "b", "c": ""}
    assert template.render_relative_root_in_output == Path("{{ cookiecutter.a }}")
    assert template.render_relative_root_in_template == Path("{{ cookiecutter.a }}")


def test_get_copier_local_template():
    finder = CopierFinder()
    template = finder.find(str(COPIER_ONE_DIR), COPIER_ONE_DIR)
    assert template.path == COPIER_ONE_DIR
    assert template.name == COPIER_ONE_NAME
    assert template.git_url is None
    assert template.version == COPIER_ONE_VERSION
    assert template.config.defaults == {"q1": "a1", "q2": 1, "q3": None}
    assert template.render_relative_root_in_output == Path(".")
    assert template.render_relative_root_in_template == Path(".")


def test_get_copier_local_output_subdir_template():
    finder = CopierFinder()
    template = finder.find(str(COPIER_OUTPUT_SUBDIR_DIR), COPIER_OUTPUT_SUBDIR_DIR)
    assert template.path == COPIER_OUTPUT_SUBDIR_DIR
    assert template.name == COPIER_OUTPUT_SUBDIR_NAME
    assert template.git_url is None
    assert template.version == COPIER_OUTPUT_SUBDIR_VERSION
    assert template.config.defaults == {"qone": "aone", "qtwo": "atwo"}
    assert template.render_relative_root_in_output == Path(".")
    assert template.render_relative_root_in_template == Path("output")


@pytest.mark.parametrize(
    "version, expect_contents",
    [
        (COOKIECUTTER_REMOTE_VERSION_1, "{{ cookiecutter.key }}"),
        (COOKIECUTTER_REMOTE_VERSION_2, "some new header\n{{ cookiecutter.key }}"),
    ],
)
def test_get_cookiecutter_remote_template(version: str, expect_contents: str):
    finder = CookiecutterFinder()
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        local_path, name = get_local_repo_path_and_name_cloning_if_repo_url(
            COOKIECUTTER_REMOTE_URL, version, dst_folder=temp_path
        )
        template = finder.find(
            COOKIECUTTER_REMOTE_URL, local_path, version=version, name=name
        )
        assert template.path == temp_path / COOKIECUTTER_REMOTE_NAME / version
        assert template.git_url == COOKIECUTTER_REMOTE_URL
        assert template.name == COOKIECUTTER_REMOTE_NAME
        assert template.version == version
        assert template.config.defaults == {"name": "abc", "key": "value"}
        assert template.render_relative_root_in_output == Path(
            "{{ cookiecutter.name }}"
        )
        assert template.render_relative_root_in_template == Path(
            "{{ cookiecutter.name }}"
        )
        template_file = (
            template.path / "{{ cookiecutter.name }}" / "{{ cookiecutter.name }}.txt"
        )
        assert template_file.read_text() == expect_contents


@pytest.mark.parametrize(
    "version, expect_contents",
    [
        (COPIER_REMOTE_VERSION_1, "{{ question2 }}"),
        (COPIER_REMOTE_VERSION_2, "{{ question2 }}\nsome new footer"),
    ],
)
def test_get_copier_remote_template(version: str, expect_contents: str):
    finder = CopierFinder()
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        local_path, name = get_local_repo_path_and_name_cloning_if_repo_url(
            COPIER_REMOTE_URL, version, dst_folder=temp_path
        )
        template = finder.find(
            COPIER_REMOTE_URL, local_path, version=version, name=name
        )
        assert template.path == temp_path / COPIER_REMOTE_NAME / version
        assert template.git_url == COPIER_REMOTE_URL
        assert template.name == COPIER_REMOTE_NAME
        assert template.version == version
        assert template.config.defaults == {"question1": "answer1", "question2": 2.7}
        assert template.render_relative_root_in_output == Path(".")
        assert template.render_relative_root_in_template == Path("output")
        template_file = template.path / "output" / "{{ question1 }}.txt.jinja"
        assert template_file.read_text() == expect_contents


def test_multi_finder_get_cookiecutter_local_template():
    finder = MultiFinder()
    template = finder.find(str(COOKIECUTTER_ONE_DIR))
    _assert_template_matches_cookiecutter_one(template)


def test_multi_finder_get_cookiecutter_local_template_from_current_directory():
    finder = MultiFinder()
    with change_directory_to(COOKIECUTTER_ONE_DIR):
        template = finder.find(str("."))
    _assert_template_matches_cookiecutter_one(template, expect_path=Path("."))


def test_multi_finder_get_copier_local_template():
    finder = MultiFinder()
    template = finder.find(str(COPIER_ONE_DIR))
    assert template.path == COPIER_ONE_DIR
    assert template.git_url is None
    assert template.version == COPIER_ONE_VERSION
    assert template.config.defaults == {"q1": "a1", "q2": 1, "q3": None}
    assert template.render_relative_root_in_output == Path(".")
    assert template.render_relative_root_in_template == Path(".")


@pytest.mark.parametrize(
    "version, expect_contents",
    [
        (COOKIECUTTER_REMOTE_VERSION_1, "{{ cookiecutter.key }}"),
        (COOKIECUTTER_REMOTE_VERSION_2, "some new header\n{{ cookiecutter.key }}"),
    ],
)
def test_multi_finder_get_cookiecutter_remote_template(
    version: str, expect_contents: str
):
    finder = MultiFinder()
    template = finder.find(COOKIECUTTER_REMOTE_URL, version=version)
    assert template.path == GENERATED_FILES_DIR / COOKIECUTTER_REMOTE_NAME / version
    assert template.git_url == COOKIECUTTER_REMOTE_URL
    assert template.version == version
    assert template.config.defaults == {"name": "abc", "key": "value"}
    assert template.render_relative_root_in_output == Path("{{ cookiecutter.name }}")
    assert template.render_relative_root_in_template == Path("{{ cookiecutter.name }}")
    template_file = (
        template.path / "{{ cookiecutter.name }}" / "{{ cookiecutter.name }}.txt"
    )
    assert template_file.read_text() == expect_contents


@pytest.mark.parametrize(
    "version, expect_contents",
    [
        (COPIER_REMOTE_VERSION_1, "{{ question2 }}"),
        (COPIER_REMOTE_VERSION_2, "{{ question2 }}\nsome new footer"),
    ],
)
def test_multi_finder_get_copier_remote_template(version: str, expect_contents: str):
    finder = MultiFinder()
    template = finder.find(COPIER_REMOTE_URL, version=version)
    assert template.path == GENERATED_FILES_DIR / COPIER_REMOTE_NAME / version
    assert template.git_url == COPIER_REMOTE_URL
    assert template.version == version
    assert template.config.defaults == {"question1": "answer1", "question2": 2.7}
    assert template.render_relative_root_in_output == Path(".")
    assert template.render_relative_root_in_template == Path("output")
    template_file = template.path / "output" / "{{ question1 }}.txt.jinja"
    assert template_file.read_text() == expect_contents



import pytest

from flexlate.template.cookiecutter import CookiecutterTemplate
from flexlate.update.template import TemplateUpdate
from tests.config import GENERATED_FILES_DIR
from tests.fixtures.template import cookiecutter_one_template


@pytest.fixture
def cookiecutter_one_update_no_data(
    cookiecutter_one_template: CookiecutterTemplate,
) -> TemplateUpdate:
    return TemplateUpdate(
        template=cookiecutter_one_template,
        config_location=GENERATED_FILES_DIR / "flexlate.json",
        index=0,
    )



from typing import Optional

from git import Repo, GitCommandError

from flexlate.branch_update import (
    get_flexlate_branch_name_for_feature_branch,
    abort_merge_and_reset_flexlate_branches,
)
from flexlate.cli_utils import confirm_user
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.ext_git import (
    fast_forward_branch_without_checkout,
    temp_repo_that_pushes_to_branch,
    repo_has_merge_conflicts,
    get_branch_sha,
    delete_local_branch,
    merge_branch_into_current,
)
from flexlate.styles import (
    print_styled,
    INFO_STYLE,
    ACTION_REQUIRED_STYLE,
    styled,
    QUESTION_STYLE,
    ALERT_STYLE,
    SUCCESS_STYLE,
)


class Merger:
    def merge_flexlate_branches(
        self,
        repo: Repo,
        branch_name: Optional[str] = None,
        delete: bool = True,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    ):
        feature_branch = branch_name or repo.active_branch.name
        flexlate_feature_merged_branch_name = (
            get_flexlate_branch_name_for_feature_branch(
                feature_branch, merged_branch_name
            )
        )
        flexlate_feature_template_branch_name = (
            get_flexlate_branch_name_for_feature_branch(
                feature_branch, template_branch_name
            )
        )

        # Save the status of the flexlate branches. We may need to roll back to this state
        # if the user aborts the merge
        current_branch = repo.active_branch
        merged_branch_sha = get_branch_sha(repo, merged_branch_name)
        template_branch_sha = get_branch_sha(repo, template_branch_name)

        print_styled(
            f"Merging {flexlate_feature_template_branch_name} to {template_branch_name}",
            INFO_STYLE,
        )
        try:
            fast_forward_branch_without_checkout(
                repo, template_branch_name, flexlate_feature_template_branch_name
            )
        except GitCommandError as e:
            if not "non-fast-forward" in str(e):
                # Got some unexpected git error, raise it
                raise e
            # Could not fast forward. Must do a merge in a temp repo and have user resolve any conflicts
            with temp_repo_that_pushes_to_branch(  # type: ignore
                repo,
                branch_name=template_branch_name,
                base_branch_name=template_branch_name,
                additional_branches=(flexlate_feature_template_branch_name,),
                copy_current_configs=False,
            ) as temp_repo:
                merge_branch_into_current(temp_repo, flexlate_feature_template_branch_name)  # type: ignore
                if repo_has_merge_conflicts(temp_repo):
                    print_styled(
                        f"Encountered merge conflicts while merging "
                        f"{flexlate_feature_template_branch_name} into {template_branch_name}",
                        INFO_STYLE,
                    )
                    print_styled(
                        f"Flexlate uses a temporary repo for this merge. "
                        f"Please resolve conflicts in {temp_repo.working_dir}",  # type: ignore
                        ACTION_REQUIRED_STYLE,
                    )
                    handled_conflicts = confirm_user(
                        styled(
                            f"Successfully handled conflicts in {temp_repo.working_dir}? n to abort",  # type: ignore
                            QUESTION_STYLE,
                        )
                    )
                    if not handled_conflicts:
                        # Have only done work in a temp dir so far, so can just abort
                        print_styled("Aborting merge.", ALERT_STYLE)
                        return
        print_styled(
            f"Successfully merged {flexlate_feature_template_branch_name} to {template_branch_name}",
            SUCCESS_STYLE,
        )

        # Now merge the merged branch in the main repo
        try:
            fast_forward_branch_without_checkout(
                repo, merged_branch_name, flexlate_feature_merged_branch_name
            )
        except GitCommandError as e:
            if not "non-fast-forward" in str(e):
                # Got some unexpected git error, raise it
                raise e
            # Could not fast forward. Must do a merge and have user resolve any conflicts
            merged_branch = repo.branches[merged_branch_name]  # type: ignore
            merged_branch.checkout()
            # First merge the template branch, so that user won't have to resolve the same conflict
            # from there again
            merge_branch_into_current(repo, template_branch_name)
            # Now merge the feature merged branch to get new changes from feature branch
            merge_branch_into_current(repo, flexlate_feature_merged_branch_name)
            if repo_has_merge_conflicts(repo):
                print_styled(
                    f"Encountered merge conflicts while merging "
                    f"{flexlate_feature_template_branch_name} into {template_branch_name}",
                    INFO_STYLE,
                )
                print_styled(f"Please resolve the conflicts", ACTION_REQUIRED_STYLE)
                handled_conflicts = confirm_user(
                    styled("Successfully handled conflicts? n to abort", QUESTION_STYLE)
                )
                if not handled_conflicts:
                    # Time ot abort, but need to reset the state of the branches
                    print_styled("Aborting merge.", ALERT_STYLE)
                    abort_merge_and_reset_flexlate_branches(
                        repo,
                        current_branch,
                        merged_branch_sha=merged_branch_sha,
                        template_branch_sha=template_branch_sha,
                        merged_branch_name=merged_branch_name,
                        template_branch_name=template_branch_name,
                    )

                    return
            current_branch.checkout()

        print_styled(
            f"Successfully merged {flexlate_feature_merged_branch_name} to {merged_branch_name}",
            SUCCESS_STYLE,
        )

        if not delete:
            return

        # Handle delete
        print_styled(
            f"Deleting flexlate feature branches {flexlate_feature_template_branch_name} and {flexlate_feature_merged_branch_name}",
            INFO_STYLE,
        )
        delete_local_branch(repo, flexlate_feature_template_branch_name)
        delete_local_branch(repo, flexlate_feature_merged_branch_name)
        print_styled(f"Successfully deleted flexlate feature branches", SUCCESS_STYLE)



from flexlate.render.multi import MultiRenderer
from flexlate.render.renderable import Renderable
from flexlate.render.specific.cookiecutter import CookiecutterRenderer
from flexlate.render.specific.copier import CopierRenderer
from tests.config import GENERATED_FILES_DIR
from tests.dirutils import wipe_generated_folder
from tests.fileutils import (
    cookiecutter_one_generated_text_content,
    cookiecutter_two_generated_text_content,
)
from tests.fixtures.template import *
from tests.fixtures.renderable import *


@pytest.fixture(autouse=True)
def before_each():
    wipe_generated_folder()


def test_render_cookiecutter_with_defaults(
    cookiecutter_one_renderable: Renderable,
):
    renderer = CookiecutterRenderer()
    data = renderer.render(cookiecutter_one_renderable, no_input=True)
    assert data == {"a": "b", "c": ""}
    assert cookiecutter_one_generated_text_content() == "b"


def test_render_copier_with_defaults(copier_one_renderable: Renderable):
    renderer = CopierRenderer()
    data = renderer.render(copier_one_renderable, no_input=True)
    assert data == {"q1": "a1", "q2": 1, "q3": None}
    rendered_path = GENERATED_FILES_DIR / "a1.txt"
    assert rendered_path.read_text() == "1"


def test_render_copier_subdir_with_defaults(
    copier_output_subdir_renderable: Renderable,
):
    renderer = CopierRenderer()
    renderable = copier_output_subdir_renderable
    data = renderer.render(renderable, no_input=True)
    assert data == {"qone": "aone", "qtwo": "atwo"}
    rendered_path = GENERATED_FILES_DIR / "aone.txt"
    assert rendered_path.read_text() == "atwo"
    should_not_exist_path = GENERATED_FILES_DIR / "not-rendered.txt"
    assert not should_not_exist_path.exists()


def test_render_cookiecutter_with_data(
    cookiecutter_one_renderable: Renderable,
):
    renderer = CookiecutterRenderer()
    cookiecutter_one_renderable.data = {"c": "something"}
    data = renderer.render(
        cookiecutter_one_renderable,
        no_input=True,
    )
    assert data == {"a": "b", "c": "something"}
    assert cookiecutter_one_generated_text_content() == "bsomething"


def test_render_copier_with_data(copier_one_renderable: Renderable):
    renderer = CopierRenderer()
    copier_one_renderable.data = {"q2": 2, "q3": "a3"}
    data = renderer.render(copier_one_renderable, no_input=True)
    assert data == {"q1": "a1", "q2": 2, "q3": "a3"}
    rendered_path = GENERATED_FILES_DIR / "a1.txt"
    assert rendered_path.read_text() == "2"


def test_render_multi_with_defaults(cookiecutter_local_renderables: List[Renderable]):
    renderer = MultiRenderer()
    data = renderer.render(
        cookiecutter_local_renderables, project_root=GENERATED_FILES_DIR, no_input=True
    )
    assert data == [{"a": "b", "c": ""}, {"a": "b", "d": "e"}]
    assert cookiecutter_one_generated_text_content() == "b"
    assert cookiecutter_two_generated_text_content() == "e"


def test_render_multi_with_copier_defaults(copier_one_renderable: Renderable):
    renderer = MultiRenderer()
    data = renderer.render(
        [copier_one_renderable], project_root=GENERATED_FILES_DIR, no_input=True
    )
    assert data == [{"q1": "a1", "q2": 1, "q3": None}]
    rendered_path = GENERATED_FILES_DIR / "a1.txt"
    assert rendered_path.read_text() == "1"


def test_render_multi_with_data(
    cookiecutter_local_renderables: List[Renderable], copier_one_renderable: Renderable
):
    renderer = MultiRenderer()
    cookiecutter_local_renderables[0].data = {"a": "z", "c": "something"}
    cookiecutter_local_renderables[1].data = {"a": "z", "d": "f"}
    copier_one_renderable.data = {"q2": 2, "q3": "a3"}
    data = renderer.render(
        [*cookiecutter_local_renderables, copier_one_renderable],
        project_root=GENERATED_FILES_DIR,
        no_input=True,
    )
    assert data == [
        {"a": "z", "c": "something"},
        {"a": "z", "d": "f"},
        {"q1": "a1", "q2": 2, "q3": "a3"},
    ]
    assert cookiecutter_one_generated_text_content(folder="z") == "zsomething"
    assert cookiecutter_two_generated_text_content(folder="z") == "f"
    copier_rendered_path = GENERATED_FILES_DIR / "a1.txt"
    assert copier_rendered_path.read_text() == "2"


def test_render_multi_with_overlap(
    cookiecutter_one_renderable: Renderable,
):
    renderer = MultiRenderer()

    data = renderer.render(
        [
            cookiecutter_one_renderable.copy(update=dict(data={"c": "something"})),
            cookiecutter_one_renderable.copy(update=dict(data={"c": "something else"})),
        ],
        no_input=True,
    )
    assert data == [{"a": "b", "c": "something"}, {"a": "b", "c": "something else"}]
    assert cookiecutter_one_generated_text_content() == "bsomethingbsomething else"


def test_render_string_local_cookiecutter(
    cookiecutter_one_renderable: Renderable,
):
    renderer = CookiecutterRenderer()
    output = renderer.render_string(
        "{{ cookiecutter.a }} works", cookiecutter_one_renderable
    )
    assert output == "b works"



import json
import os.path
from pathlib import Path
from typing import Optional

from cookiecutter.find import find_template

from flexlate.finder.specific.base import TemplateFinder
from flexlate.finder.specific.git import (
    get_version_from_source_path,
    get_git_url_from_source_path,
)
from flexlate.template.cookiecutter import CookiecutterTemplate
from flexlate.template_config.cookiecutter import CookiecutterConfig


class CookiecutterFinder(TemplateFinder[CookiecutterTemplate]):
    def find(
        self, path: str, local_path: Path, **template_kwargs
    ) -> CookiecutterTemplate:
        git_version: Optional[str] = None
        if "version" in template_kwargs:
            git_version = template_kwargs.pop("version")
        config = self.get_config(local_path)
        version = get_version_from_source_path(path, local_path) or git_version
        git_url = get_git_url_from_source_path(path, template_kwargs)
        template_source_path = git_url if git_url else path
        absolute_template_dir = find_template(local_path)
        relative_template_dir = Path(
            os.path.relpath(absolute_template_dir, local_path.resolve())
        )
        return CookiecutterTemplate(
            config,
            local_path,
            relative_template_dir,
            version=version,
            git_url=git_url,
            template_source_path=template_source_path,
            **template_kwargs
        )

    def get_config(self, directory: Path) -> CookiecutterConfig:
        config_path = directory / "cookiecutter.json"
        data = json.loads(config_path.read_text())
        return CookiecutterConfig(data)

    def matches_template_type(self, path: Path) -> bool:
        return (path / "cookiecutter.json").exists()



from dataclasses import dataclass
from enum import Enum
from typing import List, Final

import pytest

from flexlate.main import Flexlate
from tests.integration.cli_stub import CLIStubFlexlate, ExceptionHandling


class FlexlateType(str, Enum):
    APP = "app"
    CLI = "cli"


@dataclass
class FlexlateFixture:
    flexlate: Flexlate
    type: FlexlateType


flexlate_app_fixture: Final[FlexlateFixture] = FlexlateFixture(
    flexlate=Flexlate(), type=FlexlateType.APP
)

# NOTE: If any integration tests are failing, temporarily comment the CLIStubFlexlate
#  fixture while debugging, as the CLI tests are the same but much harder to debug
flexlate_fixtures: Final[List[FlexlateFixture]] = [
    flexlate_app_fixture,
    FlexlateFixture(flexlate=CLIStubFlexlate(), type=FlexlateType.CLI),
]

flexlate_ignore_cli_exceptions_fixtures: Final[List[FlexlateFixture]] = [
    flexlate_app_fixture,
    FlexlateFixture(
        flexlate=CLIStubFlexlate(exception_handling=ExceptionHandling.IGNORE),
        type=FlexlateType.CLI,
    ),
]


@pytest.fixture(scope="module", params=flexlate_fixtures)
def flexlates(request) -> FlexlateFixture:
    return request.param


@pytest.fixture(scope="module", params=flexlate_ignore_cli_exceptions_fixtures)
def flexlates_ignore_cli_exceptions(request) -> FlexlateFixture:
    return request.param



from typing import Optional

import pytest
from git import Repo

from flexlate.checker import Checker
from flexlate.config import FlexlateConfig
from flexlate.exc import TemplateNotRegisteredException
from flexlate.template.copier import CopierTemplate
from tests.config import COOKIECUTTER_REMOTE_NAME, COPIER_ONE_NAME
from tests.fixtures.templated_repo import *
from tests.fixtures.template import *
from tests.fixtures.transaction import *


@pytest.mark.parametrize("names", [None, [COOKIECUTTER_REMOTE_NAME], [COPIER_ONE_NAME]])
def test_check_for_remote_template_update(
    names: Optional[List[str]],
    repo_with_cookiecutter_remote_version_one_template_source_and_no_target_version: Repo,
    copier_one_template: CopierTemplate,
    add_source_transaction: FlexlateTransaction,
):
    repo = (
        repo_with_cookiecutter_remote_version_one_template_source_and_no_target_version
    )
    template = copier_one_template
    checker = Checker()
    adder = Adder()

    # Add a source that is already up to date
    adder.add_template_source(
        repo, template, add_source_transaction, out_root=GENERATED_REPO_DIR
    )

    new_versions = checker.find_new_versions_for_template_sources(
        names=names, project_root=GENERATED_REPO_DIR
    ).update_version_dict
    if names is None or names == [COOKIECUTTER_REMOTE_NAME]:
        assert new_versions == {COOKIECUTTER_REMOTE_NAME: COOKIECUTTER_REMOTE_VERSION_2}
    else:
        assert new_versions == {}


def test_check_for_update_with_no_template_sources(
    repo_with_placeholder_committed: Repo,
):
    checker = Checker()
    new_versions = checker.find_new_versions_for_template_sources(
        project_root=GENERATED_REPO_DIR
    ).update_version_dict
    assert new_versions == {}


def test_check_for_update_template_that_does_not_exist(
    repo_with_placeholder_committed: Repo,
):
    checker = Checker()
    with pytest.raises(TemplateNotRegisteredException):
        checker.find_new_versions_for_template_sources(
            names=["some-fake-template"], project_root=GENERATED_REPO_DIR
        )



import shutil
import tempfile
from copy import deepcopy
from pathlib import Path
from typing import List, TypedDict, Union

import pytest

from flexlate.finder.specific.cookiecutter import CookiecutterFinder
from flexlate.finder.specific.copier import CopierFinder
from flexlate.path_ops import change_directory_to
from flexlate.template.cookiecutter import CookiecutterTemplate
from flexlate.template_path import get_local_repo_path_and_name_cloning_if_repo_url
from tests.config import (
    COOKIECUTTER_ONE_DIR,
    COOKIECUTTER_TWO_DIR,
    COOKIECUTTER_REMOTE_URL,
    COOKIECUTTER_REMOTE_VERSION_2,
    COOKIECUTTER_REMOTE_VERSION_1,
    COPIER_ONE_DIR,
    COPIER_REMOTE_VERSION_2,
    COPIER_REMOTE_VERSION_1,
    COOKIECUTTER_ONE_VERSION,
    COOKIECUTTER_ONE_MODIFIED_VERSION,
    COPIER_ONE_VERSION,
    COPIER_ONE_MODIFIED_VERSION,
    COPIER_OUTPUT_SUBDIR_DIR,
    GENERATED_REPO_DIR,
    COPIER_FROM_COOKIECUTTER_ONE_DIR,
    COPIER_FROM_COOKIECUTTER_ONE_VERSION,
    COOKIECUTTER_WITH_HOOKS_DIR,
    COPIER_WITH_TASKS_DIR,
    COOKIECUTTER_WITH_HOOKS_VERSION,
    COOKIECUTTER_WITH_HOOKS_MODIFIED_VERSION,
    COPIER_WITH_TASKS_VERSION,
    COPIER_WITH_TASKS_MODIFIED_VERSION,
)


class CookiecutterRemoteTemplateData(TypedDict):
    name: str
    key: str


def get_header_for_cookiecutter_remote_template(version: str) -> str:
    if version == COOKIECUTTER_REMOTE_VERSION_2:
        return "some new header\n"
    elif version == COOKIECUTTER_REMOTE_VERSION_1:
        return ""
    else:
        raise ValueError(f"unknown cookiecutter remote version {version}")


def get_footer_for_copier_remote_template(version: str) -> str:
    if version == COPIER_REMOTE_VERSION_2:
        return "\nsome new footer"
    elif version == COPIER_REMOTE_VERSION_1:
        return ""
    else:
        raise ValueError(f"unknown copier remote version {version}")


def get_footer_for_cookiecutter_local_template(version: str) -> str:
    if version in (
        COOKIECUTTER_ONE_MODIFIED_VERSION,
        COOKIECUTTER_WITH_HOOKS_MODIFIED_VERSION,
    ):
        return " and extra"
    elif version in (
        COOKIECUTTER_ONE_VERSION,
        COOKIECUTTER_WITH_HOOKS_VERSION,
        COPIER_FROM_COOKIECUTTER_ONE_VERSION,
    ):
        return ""
    else:
        raise ValueError(f"unknown cookiecutter local version {version}")


def get_footer_for_copier_local_template(version: str) -> str:
    if version in (COPIER_ONE_MODIFIED_VERSION, COPIER_WITH_TASKS_MODIFIED_VERSION):
        return " and some footer"
    elif version in (COPIER_ONE_VERSION, COPIER_WITH_TASKS_VERSION):
        return ""
    else:
        raise ValueError(f"unknown copier local version {version}")


@pytest.fixture
def cookiecutter_one_template() -> CookiecutterTemplate:
    finder = CookiecutterFinder()
    yield finder.find(str(COOKIECUTTER_ONE_DIR), COOKIECUTTER_ONE_DIR)


@pytest.fixture
def cookiecutter_one_template_in_repo() -> CookiecutterTemplate:
    # NOTE: Must be used in conjunction with an in_repo templated repo because
    # these changes only affect the template and not the generated repo
    new_folder = GENERATED_REPO_DIR / "templates"
    new_folder.mkdir(parents=True)
    new_path = new_folder / COOKIECUTTER_ONE_DIR.name
    shutil.copytree(COOKIECUTTER_ONE_DIR, new_path)
    finder = CookiecutterFinder()
    with change_directory_to(GENERATED_REPO_DIR):
        relative_path = new_path.relative_to(GENERATED_REPO_DIR)
        template = finder.find(str(relative_path), relative_path)
    yield template


@pytest.fixture
def cookiecutter_two_template() -> CookiecutterTemplate:
    finder = CookiecutterFinder()
    yield finder.find(str(COOKIECUTTER_TWO_DIR), COOKIECUTTER_TWO_DIR)


@pytest.fixture
def cookiecutter_with_hooks_template() -> CookiecutterTemplate:
    finder = CookiecutterFinder()
    yield finder.find(str(COOKIECUTTER_WITH_HOOKS_DIR), COOKIECUTTER_WITH_HOOKS_DIR)


@pytest.fixture
def copier_one_template() -> CookiecutterTemplate:
    finder = CopierFinder()
    yield finder.find(str(COPIER_ONE_DIR), COPIER_ONE_DIR)


@pytest.fixture
def copier_output_subdir_template() -> CookiecutterTemplate:
    finder = CopierFinder()
    yield finder.find(str(COPIER_OUTPUT_SUBDIR_DIR), COPIER_OUTPUT_SUBDIR_DIR)


@pytest.fixture
def copier_with_tasks_template() -> CookiecutterTemplate:
    finder = CopierFinder()
    yield finder.find(str(COPIER_WITH_TASKS_DIR), COPIER_WITH_TASKS_DIR)


@pytest.fixture()
def cookiecutter_remote_template() -> CookiecutterTemplate:
    finder = CookiecutterFinder()
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        local_path, name = get_local_repo_path_and_name_cloning_if_repo_url(
            COOKIECUTTER_REMOTE_URL, dst_folder=temp_path
        )
        template = finder.find(COOKIECUTTER_REMOTE_URL, local_path, name=name)
        # This would normally happen when loading template from source in config
        template.template_source_path = COOKIECUTTER_REMOTE_URL
        yield template


@pytest.fixture()
def cookiecutter_remote_version_one_template() -> CookiecutterTemplate:
    finder = CookiecutterFinder()
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        local_path, name = get_local_repo_path_and_name_cloning_if_repo_url(
            COOKIECUTTER_REMOTE_URL, COOKIECUTTER_REMOTE_VERSION_1, dst_folder=temp_path
        )
        template = finder.find(
            COOKIECUTTER_REMOTE_URL,
            local_path,
            name=name,
            version=COOKIECUTTER_REMOTE_VERSION_1,
        )
        # This would normally happen when loading template from source in config
        template.template_source_path = COOKIECUTTER_REMOTE_URL
        yield template


@pytest.fixture
def cookiecutter_local_templates(
    cookiecutter_one_template: CookiecutterTemplate,
    cookiecutter_two_template: CookiecutterTemplate,
) -> List[CookiecutterTemplate]:
    yield [cookiecutter_one_template, cookiecutter_two_template]


def modify_cookiecutter_one(root: Union[str, Path]):
    text_path = Path(root) / "{{ cookiecutter.a }}" / "text.txt"
    text_path.write_text("{{ cookiecutter.a }}{{ cookiecutter.c }} and extra")


def modify_copier_one(root: Union[str, Path]):
    text_path = Path(root) / "{{ q1 }}.txt.jinja"
    text_path.write_text("{{ q2 }} and some footer")


def modify_copier_with_tasks(root: Union[str, Path]):
    text_path = Path(root) / "{{ q1 }}.txt.jinja"
    text_path.write_text("{{ q1 }} and some footer")


def modify_cookiecutter_one_to_be_copier(root: Union[str, Path]):
    shutil.rmtree(root)
    shutil.copytree(COPIER_FROM_COOKIECUTTER_ONE_DIR, root)


@pytest.fixture
def cookiecutter_one_modified_template(
    cookiecutter_one_template: CookiecutterTemplate,
) -> CookiecutterTemplate:
    template = deepcopy(cookiecutter_one_template)
    with tempfile.TemporaryDirectory() as directory:
        out_path = Path(directory)
        shutil.copytree(template.path, out_path, dirs_exist_ok=True)
        template.path = out_path
        modify_cookiecutter_one(out_path)
        # Update version
        template.version = template.folder_hash
        yield template



from pathlib import Path
from typing import Optional

from git import Repo

from flexlate.add_mode import AddMode
from flexlate.adder import Adder
from flexlate.config_manager import ConfigManager
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.render.multi import MultiRenderer
from flexlate.styles import print_styled, INFO_STYLE, SUCCESS_STYLE
from flexlate.template.base import Template
from flexlate.template_data import TemplateData
from flexlate.transactions.transaction import FlexlateTransaction
from flexlate.update.main import Updater
from flexlate.user_config_manager import UserConfigManager


class Bootstrapper:
    def bootstrap_flexlate_init_from_existing_template(
        self,
        repo: Repo,
        template: Template,
        transaction: FlexlateTransaction,
        target_version: Optional[str] = None,
        data: Optional[TemplateData] = None,
        default_add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        remote: str = "origin",
        no_input: bool = False,
        adder: Adder = Adder(),
        config_manager: ConfigManager = ConfigManager(),
        renderer: MultiRenderer = MultiRenderer(),
        updater: Updater = Updater(),
        user_config_manager: UserConfigManager = UserConfigManager(),
    ):
        if repo.working_dir is None:
            raise ValueError("repo working dir must not be None")
        project_root = Path(repo.working_dir)

        print_styled(
            f"Bootstrapping {project_root} into a Flexlate project based off the template from {template.template_source_path}",
            INFO_STYLE,
        )

        adder.init_project_and_add_to_branches(
            repo,
            default_add_mode=default_add_mode,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
            remote=remote,
            config_manager=config_manager,
        )
        adder.add_template_source(
            repo,
            template,
            transaction,
            target_version=target_version,
            out_root=project_root,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
            base_merged_branch_name=base_merged_branch_name,
            base_template_branch_name=base_template_branch_name,
            remote=remote,
            config_manager=config_manager,
        )
        adder.apply_template_and_add(
            repo,
            template,
            transaction,
            data=data,
            out_root=project_root,
            add_mode=AddMode.LOCAL,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
            base_merged_branch_name=base_merged_branch_name,
            base_template_branch_name=base_template_branch_name,
            no_input=no_input,
            remote=remote,
            config_manager=config_manager,
            updater=updater,
            renderer=renderer,
        )
        if target_version is not None:
            # Remove peg to specific version if one was passed
            user_config_manager.update_template_source_target_version(
                template.name,
                None,
                repo,
                transaction,
                project_path=project_root,
                add_mode=AddMode.LOCAL,
                merged_branch_name=merged_branch_name,
                template_branch_name=template_branch_name,
                base_merged_branch_name=base_merged_branch_name,
                base_template_branch_name=base_template_branch_name,
                remote=remote,
                config_manager=config_manager,
            )

        print_styled(
            f"Successfully bootstrapped {project_root} into a Flexlate project based off the template from {template.template_source_path}",
            SUCCESS_STYLE,
        )



from typing import Optional, Sequence

from git import Repo

from flexlate.branch_update import get_flexlate_branch_name_for_feature_branch
from flexlate.constants import DEFAULT_TEMPLATE_BRANCH_NAME, DEFAULT_MERGED_BRANCH_NAME
from flexlate.ext_git import branch_exists, push_to_remote
from flexlate.styles import print_styled, ALERT_STYLE, INFO_STYLE, SUCCESS_STYLE


class Pusher:
    def push_main_flexlate_branches(
        self,
        repo: Repo,
        remote: str = "origin",
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    ):
        _push_branches_to_remote(
            repo, [template_branch_name, merged_branch_name], remote=remote
        )

    def push_feature_flexlate_branches(
        self,
        repo: Repo,
        feature_branch: Optional[str] = None,
        remote: str = "origin",
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    ):
        branch_name = feature_branch or repo.active_branch.name
        feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
            branch_name, merged_branch_name
        )
        feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
            branch_name, template_branch_name
        )
        _push_branches_to_remote(
            repo,
            [feature_template_branch_name, feature_merged_branch_name],
            remote=remote,
        )


def _push_branches_to_remote(
    repo: Repo, branch_names: Sequence[str], remote: str = "origin"
):
    for branch in branch_names:
        if not branch_exists(repo, branch):
            print_styled(
                f"Could not push branch {branch} as it does not exist", ALERT_STYLE
            )
            return
    and_branches = " and ".join(branch_names)
    print_styled(
        f"Pushing {and_branches} to remote {remote}",
        INFO_STYLE,
    )
    for branch in branch_names:
        push_to_remote(repo, branch, remote_name=remote)
    print_styled("Successfully pushed branches to remote", SUCCESS_STYLE)



import os.path

from flexlate.main import Flexlate
from flexlate.template.copier import CopierTemplate
from flexlate.template.types import TemplateType
from tests.config import GENERATED_FILES_DIR
from tests.fixtures.templated_repo import *

fxt = Flexlate()


def test_add_template_source_from_current_path(
    repo_with_placeholder_committed: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
):
    template = cookiecutter_one_template

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        with change_directory_to(COOKIECUTTER_ONE_DIR):
            fxt.add_template_source(".", template_root=GENERATED_REPO_DIR)

    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 0
    assert len(config.template_sources) == 1
    source = config.template_sources[0]
    assert source.name == template.name
    assert source.path == os.path.relpath(COOKIECUTTER_ONE_DIR, GENERATED_REPO_DIR)
    assert source.version == template.version
    assert source.type == TemplateType.COOKIECUTTER
    assert source.render_relative_root_in_output == Path("{{ cookiecutter.a }}")
    assert source.render_relative_root_in_template == Path("{{ cookiecutter.a }}")


def test_init_from_current_path_cookiecutter(
    cookiecutter_one_template: CookiecutterTemplate,
):
    template = cookiecutter_one_template
    project_dir = GENERATED_FILES_DIR / "b"

    with change_directory_to(template.path):
        fxt.init_project_from(".", path=GENERATED_FILES_DIR, no_input=True)

    content_path = project_dir / "text.txt"
    content = content_path.read_text()
    assert content == "b"

    config_path = project_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    source = config.template_sources[0]
    assert source.name == template.name
    assert source.path == "../../input_files/templates/cookiecutters/one"
    assert source.version == template.version
    assert source.type == TemplateType.COOKIECUTTER
    assert source.render_relative_root_in_output == Path("{{ cookiecutter.a }}")
    assert source.render_relative_root_in_template == Path("{{ cookiecutter.a }}")
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"a": "b", "c": ""}
    assert at.root == Path("..")
    assert at.add_mode == AddMode.LOCAL


def test_init_from_current_path_copier(
    copier_one_template: CopierTemplate,
):
    template = copier_one_template
    project_dir = GENERATED_FILES_DIR / "project"

    with change_directory_to(template.path):
        fxt.init_project_from(".", path=GENERATED_FILES_DIR, no_input=True)

    content_path = project_dir / "a1.txt"
    content = content_path.read_text()
    assert content == "1"

    readme_path = project_dir / "README.md"
    assert readme_path.read_text() == "some existing content"

    config_path = project_dir / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    source = config.template_sources[0]
    assert source.name == template.name
    assert source.path == "../../input_files/templates/copiers/one"
    assert source.version == template.version
    assert source.type == TemplateType.COPIER
    assert source.render_relative_root_in_output == Path(".")
    assert source.render_relative_root_in_template == Path(".")
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.name == template.name
    assert at.version == template.version
    assert at.data == {"q1": "a1", "q2": 1, "q3": None}
    assert at.root == Path(".")
    assert at.add_mode == AddMode.LOCAL



from pathlib import Path
from typing import Optional

from flexlate.add_mode import AddMode
from flexlate.config import FlexlateConfig, FlexlateProjectConfig
from flexlate.template_data import TemplateData
from tests.config import (
    GENERATED_REPO_DIR,
    COOKIECUTTER_REMOTE_VERSION_2,
    COOKIECUTTER_REMOTE_NAME,
    COOKIECUTTER_REMOTE_URL,
    GENERATED_FILES_DIR,
)
from tests.fixtures.template import (
    get_header_for_cookiecutter_remote_template,
    get_footer_for_copier_remote_template,
    get_footer_for_cookiecutter_local_template,
    get_footer_for_copier_local_template,
    CookiecutterRemoteTemplateData,
)
from tests.integration.fixtures.template_source import (
    TemplateSourceType,
    COOKIECUTTER_REMOTE_DEFAULT_EXPECT_PATH,
    TemplateSourceFixture,
)


def assert_project_files_are_correct(
    root: Path = GENERATED_REPO_DIR,
    expect_data: Optional[TemplateData] = None,
    version: str = COOKIECUTTER_REMOTE_VERSION_2,
    template_source_type: TemplateSourceType = TemplateSourceType.COOKIECUTTER_REMOTE,
):
    default_data = _get_default_data(template_source_type)
    data: TemplateData = expect_data or default_data

    if template_source_type == TemplateSourceType.COOKIECUTTER_REMOTE:
        out_path = root / data["name"] / f"{data['name']}.txt"
        header = get_header_for_cookiecutter_remote_template(version)
        expect_content = f"{header}{data['key']}"
    elif template_source_type == TemplateSourceType.COPIER_REMOTE:
        out_path = root / f"{data['question1']}.txt"
        footer = get_footer_for_copier_remote_template(version)
        expect_content = f"{data['question2']}{footer}"
    elif template_source_type == TemplateSourceType.COOKIECUTTER_LOCAL:
        out_path = root / data["a"] / "text.txt"
        footer = get_footer_for_cookiecutter_local_template(version)
        expect_content = f"{data['a']}{data['c']}{footer}"
    elif template_source_type == TemplateSourceType.COPIER_LOCAL:
        out_path = root / f"{data['q1']}.txt"
        footer = get_footer_for_copier_local_template(version)
        expect_content = f"{data['q2']}{footer}"
    elif template_source_type == TemplateSourceType.COPIER_WITH_TASKS:
        out_path = root / f"{data['q1']}.txt"
        footer = get_footer_for_copier_local_template(version)
        expect_content = f"{data['q1']}{footer}"
    else:
        raise ValueError(f"unexpected template source type {template_source_type}")

    assert out_path.exists()
    content = out_path.read_text()
    assert content == expect_content


def assert_project_files_do_not_exist(
    root: Path = GENERATED_REPO_DIR,
    expect_data: Optional[CookiecutterRemoteTemplateData] = None,
):
    data: CookiecutterRemoteTemplateData = expect_data or dict(name="abc", key="value")
    out_path = root / data["name"] / f"{data['name']}.txt"
    assert not out_path.exists()


def assert_template_sources_config_is_correct(
    config_path: Path = GENERATED_REPO_DIR / "flexlate.json",
    version: str = COOKIECUTTER_REMOTE_VERSION_2,
    target_version: Optional[str] = None,
    name: str = COOKIECUTTER_REMOTE_NAME,
    url: str = COOKIECUTTER_REMOTE_URL,
    path: str = COOKIECUTTER_REMOTE_DEFAULT_EXPECT_PATH,
    render_relative_root_in_output: Path = Path("."),
    render_relative_root_in_template: Path = Path("."),
    num_template_sources: int = 1,
    template_source_index: int = 0,
):
    assert config_path.exists()
    config = FlexlateConfig.load(config_path)
    # Template source
    assert len(config.template_sources) == num_template_sources
    template_source = config.template_sources[template_source_index]
    assert template_source.name == name
    assert template_source.version == version
    assert template_source.target_version == target_version
    assert template_source.git_url == url
    assert template_source.path == path
    assert (
        template_source.render_relative_root_in_output == render_relative_root_in_output
    )
    assert (
        template_source.render_relative_root_in_template
        == render_relative_root_in_template
    )


def assert_template_sources_config_is_empty(
    config_path: Path = GENERATED_REPO_DIR / "flexlate.json",
):
    assert config_path.exists()
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 0


def assert_applied_templates_config_is_correct(
    config_path: Path = GENERATED_REPO_DIR / "flexlate.json",
    expect_applied_template_root: Path = Path("."),
    expect_data: Optional[TemplateData] = None,
    version: str = COOKIECUTTER_REMOTE_VERSION_2,
    template_source_type: TemplateSourceType = TemplateSourceType.COOKIECUTTER_REMOTE,
    name: str = COOKIECUTTER_REMOTE_NAME,
    expect_add_mode: AddMode = AddMode.LOCAL,
):
    data: TemplateData = expect_data or _get_default_data(template_source_type)
    assert config_path.exists()
    config = FlexlateConfig.load(config_path)
    # Applied template
    assert len(config.applied_templates) == 1
    applied_template = config.applied_templates[0]
    assert applied_template.name == name
    assert applied_template.data == data
    assert applied_template.version == version
    assert applied_template.root == expect_applied_template_root
    assert applied_template.add_mode == expect_add_mode


def assert_applied_templates_config_is_empty(
    config_path: Path = GENERATED_REPO_DIR / "flexlate.json",
):
    assert config_path.exists()
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 0


def assert_config_is_correct(
    at_config_path: Path = GENERATED_REPO_DIR / "flexlate.json",
    ts_config_path: Optional[Path] = GENERATED_REPO_DIR / "flexlate.json",
    expect_applied_template_root: Path = Path("."),
    expect_data: Optional[CookiecutterRemoteTemplateData] = None,
    version: str = COOKIECUTTER_REMOTE_VERSION_2,
    target_version: Optional[str] = None,
    template_source_type: TemplateSourceType = TemplateSourceType.COOKIECUTTER_REMOTE,
    name: str = COOKIECUTTER_REMOTE_NAME,
    url: str = COOKIECUTTER_REMOTE_URL,
    path: str = COOKIECUTTER_REMOTE_DEFAULT_EXPECT_PATH,
    render_relative_root_in_output: Path = Path("{{ cookiecutter.name }}"),
    render_relative_root_in_template: Path = Path("{{ cookiecutter.name }}"),
    expect_add_mode: AddMode = AddMode.LOCAL,
    num_template_sources: int = 1,
    template_source_index: int = 0,
):
    assert_applied_templates_config_is_correct(
        at_config_path,
        expect_applied_template_root,
        expect_data=expect_data,
        version=version,
        template_source_type=template_source_type,
        name=name,
        expect_add_mode=expect_add_mode,
    )
    assert_template_sources_config_is_correct(
        ts_config_path,
        version=version,
        target_version=target_version,
        name=name,
        url=url,
        path=path,
        render_relative_root_in_output=render_relative_root_in_output,
        render_relative_root_in_template=render_relative_root_in_template,
        num_template_sources=num_template_sources,
        template_source_index=template_source_index,
    )


def assert_project_config_is_correct(
    path: Path = GENERATED_REPO_DIR / "flexlate-project.json",
    user: bool = False,
    add_mode: AddMode = AddMode.LOCAL,
    project_folder_name: str = "project",
    project_containing_folder: Path = GENERATED_FILES_DIR,
):
    expect_project_path = project_containing_folder / project_folder_name
    project_config = FlexlateProjectConfig.load(path)
    assert len(project_config.projects) == 1
    project = project_config.projects[0]
    if user:
        assert project.path == project.path.absolute()
        assert project.path == expect_project_path
    else:
        assert project.path != project.path.absolute()
        assert (path.parent / project.path).absolute() == expect_project_path
    assert project.default_add_mode == add_mode


def _get_default_data(template_source_type: TemplateSourceType) -> TemplateData:
    if template_source_type == TemplateSourceType.COOKIECUTTER_REMOTE:
        return dict(name="abc", key="value")
    elif template_source_type == TemplateSourceType.COPIER_REMOTE:
        return dict(question1="answer1", question2=2.7)
    elif template_source_type == TemplateSourceType.COOKIECUTTER_LOCAL:
        return dict(a="b", c="")
    elif template_source_type == TemplateSourceType.COPIER_LOCAL:
        return dict(q1="a1", q2=1, q3=None)
    elif template_source_type == TemplateSourceType.COPIER_WITH_TASKS:
        return dict(q1="a1")
    else:
        raise ValueError(f"unexpected template source type {template_source_type}")


def assert_root_template_source_output_is_correct(
    template_source: TemplateSourceFixture,
    after_version_update: bool = False,
    after_data_update: bool = False,
    target_version: Optional[str] = None,
    num_template_sources: int = 1,
    template_source_index: int = 0,
):
    version = (
        template_source.version_2 if after_version_update else template_source.version_1
    )
    input_data = (
        template_source.update_input_data
        if after_data_update
        else template_source.input_data
    )
    at_config_path = (
        GENERATED_REPO_DIR
        / template_source.evaluated_render_relative_root_in_output_creator(input_data)
        / "flexlate.json"
    )
    assert_project_files_are_correct(
        expect_data=input_data,
        version=version,
        template_source_type=template_source.type,
    )
    assert_config_is_correct(
        at_config_path=at_config_path,
        expect_applied_template_root=template_source.expect_local_applied_template_path,
        expect_data=input_data,
        version=version,
        target_version=target_version,
        template_source_type=template_source.type,
        name=template_source.name,
        url=template_source.url,
        path=template_source.path,
        render_relative_root_in_output=template_source.render_relative_root_in_output,
        render_relative_root_in_template=template_source.render_relative_root_in_template,
        num_template_sources=num_template_sources,
        template_source_index=template_source_index,
    )



import shutil
import tempfile
from collections import ChainMap
from pathlib import Path
from typing import Sequence, Final, Any, Dict

import yaml
from copier import copy_local
from copier.config.factory import verify_minimum_version, filter_config
from copier.config.objects import ConfigData, EnvOps
from copier.config.user_data import load_config_data, query_user_data

from flexlate.render.renderable import Renderable
from flexlate.render.specific.base import SpecificTemplateRenderer
from flexlate.template.copier import CopierTemplate
from flexlate.template.types import TemplateType
from flexlate.template_data import TemplateData

exclude_copier_keys: Final[Sequence[str]] = ("now", "make_secret", "_folder_name")


class CopierRenderer(SpecificTemplateRenderer[CopierTemplate]):
    _template_cls = CopierTemplate
    _template_type = TemplateType.COPIER

    def render(
        self,
        renderable: Renderable[CopierTemplate],
        no_input: bool = False,
    ) -> TemplateData:
        template = renderable.template
        conf = _make_config_by_adding_defaults_then_prompting_user(
            str(template.path),
            str(renderable.out_root),
            data=renderable.data,
            no_input=no_input,
        )
        copy_local(conf=conf)
        return _extract_template_data_from_copier_config(conf)

    def render_string(
        self,
        string: str,
        renderable: Renderable[CopierTemplate],
    ) -> str:
        template = renderable.template

        with tempfile.TemporaryDirectory() as temp_template_dir:
            copier_config_path = _find_copier_yaml_path(template.path)
            with open(copier_config_path, "r") as f:
                copier_config = yaml.safe_load(f)
            _modify_copier_config_yaml_to_remove_tasks(copier_config)
            config_out_path = Path(temp_template_dir) / copier_config_path.name
            config_out_path.write_text(yaml.dump(copier_config))
            temp_template_file_path = (
                Path(temp_template_dir)
                / template.render_relative_root_in_template
                / "temp.txt"
            )
            if not temp_template_file_path.parent.exists():
                temp_template_file_path.parent.mkdir(parents=True)
            temp_template_file_path.write_text(string)

            with tempfile.TemporaryDirectory() as temp_output_dir:
                conf = _make_config_by_adding_defaults_then_prompting_user(
                    temp_template_dir,
                    temp_output_dir,
                    data=renderable.data,
                    no_input=True,
                )
                copy_local(conf=conf)
                temp_output_file_path = Path(temp_output_dir) / "temp.txt"
                output = temp_output_file_path.read_text()
        return output


def _modify_copier_config_yaml_to_remove_tasks(
    copier_config: Dict[str, Any]
) -> Dict[str, Any]:
    if "_tasks" in copier_config:
        del copier_config["_tasks"]
    return copier_config


def _find_copier_yaml_path(template_root: Path) -> Path:
    copier_yaml_path = template_root / "copier.yml"
    if not copier_yaml_path.exists():
        copier_yaml_path = template_root / "copier.yaml"
    if not copier_yaml_path.exists():
        raise FileNotFoundError(
            f"Could not find copier.yml or copier.yaml in {template_root}"
        )
    return copier_yaml_path


def _extract_template_data_from_copier_config(config: ConfigData) -> TemplateData:
    raw_data = dict(config.data)
    return {
        key: value for key, value in raw_data.items() if key not in exclude_copier_keys
    }


def _make_config_by_adding_defaults_then_prompting_user(
    src_path: str, dst_path: str, data: TemplateData, no_input: bool = False
) -> ConfigData:
    """
    NOTE: Adapted from copier.config.factory.make_config
    """
    init_args: Dict[str, Any] = {}
    init_args["original_src_path"] = src_path
    init_args["src_path"] = src_path
    init_args["dst_path"] = dst_path

    # Skipped logic around VCS as we are always working with a local repo by this point

    file_data = load_config_data(src_path, quiet=True)

    try:
        verify_minimum_version(file_data["_min_copier_version"])
    except KeyError:
        pass

    template_config_data, questions_data = filter_config(file_data)
    init_args["data_from_template_defaults"] = {
        k: v.get("default") for k, v in questions_data.items()
    }
    init_args["envops"] = EnvOps(**template_config_data.get("envops", {}))

    init_args["data_from_init"] = ChainMap(
        query_user_data(
            {k: v for k, v in questions_data.items() if k in data},
            {},
            data,
            False,
            init_args["envops"],
        ),
        data,
    )
    init_args["data_from_asking_user"] = query_user_data(
        questions_data,
        # This is the main difference from copier.config.factory.make_config
        # In that function, it will use data_from_init (which includes passed data)
        # as forced_answers_data, meaning it will not prompt. It uses data_from_answers_file
        # to provide the defaults for prompts. Here we are not using the answers file, so
        # we use data_from_init as last_answers_data to provide defaults and do not force
        # any answers (so all prompts appear)
        init_args["data_from_init"],
        {},
        not no_input,
        init_args["envops"],
    )
    return ConfigData(**ChainMap(init_args, template_config_data))



from pathlib import Path

import pytest

from flexlate.exc import InvalidTemplatePathException
from flexlate.template_path import (
    is_repo_url,
    is_local_template,
    get_local_repo_path_and_name_cloning_if_repo_url,
)
from tests.config import GENERATED_FILES_DIR
from tests.dirutils import wipe_generated_folder
from tests.fixtures.repo_path import (
    repo_path_fixture,
    repo_path_non_ssh_fixture,
    RepoPathFixture,
)


def test_is_repo_url(repo_path_fixture: RepoPathFixture):
    assert is_repo_url(repo_path_fixture.path) == repo_path_fixture.is_repo_url


def test_is_local_path(repo_path_fixture: RepoPathFixture):
    assert is_local_template(repo_path_fixture.path) == repo_path_fixture.is_local_path


# TODO: figure out how to test cloning SSH urls
def test_get_local_repo_path_cloning_if_repo_url(
    repo_path_non_ssh_fixture: RepoPathFixture,
):
    repo_path_fixture = repo_path_non_ssh_fixture
    wipe_generated_folder()
    if not repo_path_fixture.is_repo_url and not repo_path_fixture.is_local_path:
        # Invalid path test
        with pytest.raises(InvalidTemplatePathException):
            get_local_repo_path_and_name_cloning_if_repo_url(
                repo_path_fixture.path, GENERATED_FILES_DIR
            )
        # Path was invalid so nothing else to check, end test
        return

    # Must be valid template path, local or remote
    for version in repo_path_fixture.versions:
        local_path, name = get_local_repo_path_and_name_cloning_if_repo_url(
            repo_path_fixture.path, version, GENERATED_FILES_DIR
        )
        assert name == repo_path_fixture.name
        if repo_path_fixture.is_local_path:
            assert local_path == Path(repo_path_fixture.path)
        elif repo_path_fixture.is_repo_url:
            assert local_path == GENERATED_FILES_DIR / repo_path_fixture.name / (
                version or repo_path_fixture.default_version
            )
            repo_path_fixture.assert_was_cloned_correctly(local_path, version)
        wipe_generated_folder()



from enum import Enum
from typing import List, Final

import pytest
from git import Repo

from flexlate.ext_git import delete_local_branch
from tests.gitutils import reset_n_commits_without_checkout


class LocalBranchSituation(str, Enum):
    UP_TO_DATE = "up to date"
    DELETED = "deleted"
    OUT_OF_DATE = "out of date"

    def apply(self, repo: Repo, branch_name: str):
        apply_local_branch_situation(repo, branch_name, self)


all_local_branch_situations: Final[List[LocalBranchSituation]] = list(
    LocalBranchSituation
)


@pytest.fixture(scope="module", params=all_local_branch_situations)
def local_branch_situation(request) -> LocalBranchSituation:
    return request.param


# TODO: see if there is a better way to structure multiple parameterized fixtures
#  that actually resolve to the same value
@pytest.fixture(scope="module", params=all_local_branch_situations)
def template_branch_situation(request) -> LocalBranchSituation:
    return request.param


@pytest.fixture(scope="module", params=all_local_branch_situations)
def output_branch_situation(request) -> LocalBranchSituation:
    return request.param


def apply_local_branch_situation(
    repo: Repo, branch_name: str, situation: LocalBranchSituation
):
    if situation == LocalBranchSituation.UP_TO_DATE:
        return
    if situation == LocalBranchSituation.DELETED:
        return delete_local_branch(repo, branch_name)
    if situation == LocalBranchSituation.OUT_OF_DATE:
        return reset_n_commits_without_checkout(repo, branch_name)
    raise NotImplementedError(f"no handling for local branch situation {situation}")



import os
from pathlib import Path
from typing import List, Optional, Sequence, Dict, Any, Tuple, Union, cast

from pyappconf import BaseConfig, AppConfig, ConfigFormats
from pydantic import BaseModel, Field, validator, Extra, PrivateAttr

from flexlate.add_mode import AddMode
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.exc import (
    InvalidTemplateTypeException,
    FlexlateProjectConfigFileNotExistsException,
)
from flexlate.finder.multi import MultiFinder
from flexlate.finder.specific.base import TemplateFinder
from flexlate.finder.specific.cookiecutter import CookiecutterFinder
from flexlate.finder.specific.copier import CopierFinder
from flexlate.template.base import Template
from flexlate.template.types import TemplateType
from flexlate.template_data import TemplateData
from flexlate.template_path import get_local_repo_path_and_name_cloning_if_repo_url


class TemplateSource(BaseModel):
    name: str
    path: str
    type: TemplateType
    version: Optional[str] = None
    git_url: Optional[str] = None
    target_version: Optional[str] = None
    render_relative_root_in_output: Path = Path(".")
    render_relative_root_in_template: Path = Path(".")

    _config_file_location: Path = PrivateAttr()

    @classmethod
    def from_template(
        cls, template: Template, target_version: Optional[str] = None
    ) -> "TemplateSource":
        return cls(
            name=template.name,
            path=template.git_url
            if template.git_url is not None
            else str(template.path),
            version=template.version,
            type=template._type,
            target_version=target_version,
            git_url=template.git_url,
            render_relative_root_in_output=template.render_relative_root_in_output,
            render_relative_root_in_template=template.render_relative_root_in_template,
        )

    def to_template(
        self, version: Optional[str] = None, finder: MultiFinder = MultiFinder()
    ) -> Template:
        if self.type == TemplateType.BASE:
            raise InvalidTemplateTypeException(
                "base type is not allowed for concrete templates"
            )
        if version is None:
            version = self.version
        template = finder.find(
            self.git_url or str(self.absolute_local_path), version=version
        )
        template.name = self.name
        # TODO: Can we remove target_version from templates?
        template.target_version = self.target_version
        # Keep original template source path (may be relative), so that later when
        # updating templates, it can update the path without forcing it to be absolute
        template.template_source_path = self.path
        return template

    @property
    def update_location(self) -> Union[str, Path]:
        return self.git_url or self.absolute_local_path

    @property
    def is_local_template(self) -> bool:
        return self.git_url is None

    @property
    def absolute_local_path(self) -> Path:
        if Path(self.path).is_absolute():
            return Path(self.path)
        else:
            # Convert to absolute path
            return (self._config_file_location.parent / Path(self.path)).resolve()


class AppliedTemplateConfig(BaseModel):
    name: str
    data: TemplateData
    version: str
    add_mode: AddMode
    root: Path = Path(".")

    _config_file_location: Path = PrivateAttr()
    _orig_root: Path = PrivateAttr()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Root may get updated externally during config loading when adjusting applied paths,
        # save the original root
        self._orig_root = self.root


class AppliedTemplateWithSource(BaseModel):
    applied_template: AppliedTemplateConfig
    source: TemplateSource

    def to_template_and_data(self) -> Tuple[Template, TemplateData]:
        return (
            self.source.to_template(version=self.applied_template.version),
            self.applied_template.data,
        )


class TemplateSourceWithTemplates(BaseModel):
    source: TemplateSource
    templates: List[Template]

    class Config:
        arbitrary_types_allowed = True


class FlexlateConfig(BaseConfig):
    template_sources: List[TemplateSource] = Field(default_factory=list)
    applied_templates: List[AppliedTemplateConfig] = Field(default_factory=list)
    _child_configs: Optional[List["FlexlateConfig"]] = None
    _settings = AppConfig(
        app_name="flexlate", default_format=ConfigFormats.JSON, config_name="flexlate"
    )

    @classmethod
    def from_dir_including_nested(
        cls, root: Path, adjust_applied_paths: bool = True
    ) -> "FlexlateConfig":
        file_name = cls._settings.config_file_name
        configs = _load_nested_configs(
            root, file_name, root, adjust_applied_paths=adjust_applied_paths
        )
        # Add user config if it exists
        if cls._settings.config_location.exists():
            configs.append(cls.load())
        return cls.from_multiple(configs)

    @classmethod
    def from_multiple(cls, configs: Sequence["FlexlateConfig"]) -> "FlexlateConfig":
        template_sources: List[TemplateSource] = []
        applied_templates: List[AppliedTemplateConfig] = []
        for conf in configs:
            template_sources.extend(conf.template_sources)
            applied_templates.extend(conf.applied_templates)
        obj = cls(
            template_sources=template_sources, applied_templates=applied_templates
        )
        obj._child_configs = list(configs)
        return obj

    @validator("template_sources")
    def template_name_must_be_unique(cls, v):
        names = [template.name for template in v]
        unique_names = set(names)
        for name in unique_names:
            if names.count(name) > 1:
                raise ValueError(f"Must have unique name. Repeated: {name}")
        return v

    @property
    def template_sources_dict(self) -> Dict[str, TemplateSource]:
        return {ts.name: ts for ts in self.template_sources}

    @property
    def child_configs(self) -> List["FlexlateConfig"]:
        return self._child_configs or []

    @property
    def empty(self) -> bool:
        return len(self.template_sources) == 0 and len(self.applied_templates) == 0

    @classmethod
    def load(cls, path: Optional[Union[str, Path]] = None) -> "FlexlateConfig":
        config = cast(FlexlateConfig, super().load(path))
        for template_source in config.template_sources:
            # Add location from which config was loaded so that later template source paths
            # can be made absolute before usage
            loaded_path = path or cls._settings.config_location
            template_source._config_file_location = Path(loaded_path)
        for applied_template in config.applied_templates:
            # Add location from which config was loaded so that it is easier to move
            # applied templates to correct location after updates
            loaded_path = path or cls._settings.config_location
            applied_template._config_file_location = Path(loaded_path)
        return config

    def save(self, serializer_kwargs: Optional[Dict[str, Any]] = None, **kwargs):
        if not self.child_configs:
            # Normal singular config, fall back to py-app-conf behavior
            return super().save(serializer_kwargs, **kwargs)
        # Parent pseudo-config holding actual child configs, save those instead
        for config in self.child_configs:
            if config.empty:
                # User must have removed all the sources and applied templates,
                # therefore we don't want the config file anymore
                if config.settings.config_location.exists():
                    os.remove(config.settings.config_location)
                continue

            # Normal case, non-empty config, save it
            config.save(serializer_kwargs, **kwargs)

    class Config:
        extra = Extra.allow


def _load_nested_configs(
    root: Path, file_name: str, orig_root: Path, adjust_applied_paths: bool = True
) -> List["FlexlateConfig"]:
    configs: List["FlexlateConfig"] = []
    path = root / file_name
    if path.exists():
        relative_path = root.relative_to(orig_root)
        config = FlexlateConfig.load(path)
        # Because we are combining configs, need to update the root for the applied templates
        if adjust_applied_paths:
            for applied_template in config.applied_templates:
                applied_template.root = relative_path / applied_template.root
        configs.append(config)
    for folder_or_file in root.iterdir():
        if folder_or_file.is_dir():
            # Check to see if this is a nested flexlate project. If so, don't load the config,
            # instead that config should be associated with the nested project.
            if (folder_or_file / "flexlate-project.json").exists():
                continue

            configs.extend(
                _load_nested_configs(
                    folder_or_file,
                    file_name,
                    orig_root,
                    adjust_applied_paths=adjust_applied_paths,
                )
            )
    return configs


class ProjectConfig(BaseModel):
    path: Path
    default_add_mode: AddMode = AddMode.LOCAL
    merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME
    template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME
    remote: str = "origin"


class FlexlateProjectConfig(BaseConfig):
    projects: List[ProjectConfig] = Field(default_factory=list)
    _settings = AppConfig(
        app_name="flexlate",
        default_format=ConfigFormats.JSON,
        config_name="flexlate-project",
    )

    def get_project_for_path(self, path: Path = Path(".")) -> ProjectConfig:
        # Find the project root that is the closest parent to the given path
        project_closeness: List[Tuple[ProjectConfig, int]] = []
        for raw_project in self.projects:
            project = self._absolutify_project_config(raw_project)
            if project.path.absolute() == path.absolute():
                # If it is an exact match, just return it, we are done
                return project
            elif project.path.absolute() in path.absolute().parents:
                # Project path is a parent of the given path. Track the closeness
                # so that we can take the nearest parent
                project_closeness.append(
                    (project, path.absolute().parents.index(project.path.absolute()))
                )
            # Else, this path is totally unrelated, do nothing
        if len(project_closeness) == 0:
            raise FlexlateProjectConfigFileNotExistsException(
                f"could not find a project matching the path {path} from the "
                f"projects config file at {self.settings.config_location}"
            )

        # We have potentially multiple matching projects. Take the closest parent
        project_closeness.sort(key=lambda pc: pc[1])
        return project_closeness[0][0]

    def _absolutify_project_config(
        self, project_config: ProjectConfig
    ) -> ProjectConfig:
        if project_config.path.is_absolute():
            return project_config
        # Project config has a relative path, but that is relative to the config location,
        # not the current working directory
        real_path = self.settings.config_location.parent / project_config.path
        return ProjectConfig(
            path=real_path.absolute(), default_add_mode=project_config.default_add_mode
        )


if __name__ == "__main__":
    print(FlexlateConfig().to_json())



import re
from unittest.mock import patch

import pytest

from flexlate.branch_update import (
    get_flexlate_branch_name,
    get_flexlate_branch_name_for_feature_branch,
)
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.ext_git import merge_branch_into_current
from flexlate.merger import Merger
from flexlate import merger as merger_mod
from tests.fixtures.templated_repo import *
from tests.fixtures.template import *
from tests.fixtures.transaction import add_output_transaction
from tests.gitutils import (
    rename_branch,
    assert_main_commit_message_matches,
    checkout_new_branch,
)


@pytest.mark.parametrize("delete", [True, False])
def test_merge_with_non_existent_main_branches(
    delete: bool,
    repo_with_template_branch_from_cookiecutter_one_and_feature_flexlate_branches: Repo,
):
    repo = repo_with_template_branch_from_cookiecutter_one_and_feature_flexlate_branches
    feature_merged_branch_name = get_flexlate_branch_name(
        repo, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name(
        repo, DEFAULT_TEMPLATE_BRANCH_NAME
    )

    merger = Merger()
    merger.merge_flexlate_branches(repo, delete=delete)

    _assert_successful_merge(
        repo, feature_merged_branch_name, feature_template_branch_name, delete
    )


@pytest.mark.parametrize("delete", [True, False])
def test_merge_with_existing_main_branches(
    delete: bool,
    repo_with_template_branch_from_cookiecutter_one: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_template_branch_from_cookiecutter_one

    feature_branch_name = "feature"
    feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch_name, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch_name, DEFAULT_TEMPLATE_BRANCH_NAME
    )

    # Existing repo has main flexlate branches. Now do another operation with feature branch
    checkout_new_branch(repo, feature_branch_name)
    subdir = GENERATED_REPO_DIR / "subdir"
    subdir.mkdir()
    with change_directory_to(subdir):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_one_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
            merged_branch_name=feature_merged_branch_name,
            template_branch_name=feature_template_branch_name,
        )

    merger = Merger()
    merger.merge_flexlate_branches(repo, delete=delete)

    _assert_successful_merge(
        repo, feature_merged_branch_name, feature_template_branch_name, delete
    )


@pytest.mark.parametrize("delete", [True, False])
@pytest.mark.parametrize("commit_first", [True, False])
def test_merge_with_existing_main_branches_and_conflicts(
    delete: bool,
    commit_first: bool,
    repo_with_template_branch_from_cookiecutter_one: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_template_branch_from_cookiecutter_one

    adder = Adder()
    merger = Merger()

    # Existing repo has main flexlate branches. Now do another operation with feature branch
    subdir = GENERATED_REPO_DIR / "subdir"
    subdir.mkdir()
    with change_directory_to(subdir):
        for i in range(2):
            feature_branch_name = f"feature-{i + 1}"
            data = dict(a=str(i), c=str(i))
            feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
                feature_branch_name, DEFAULT_MERGED_BRANCH_NAME
            )
            feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
                feature_branch_name, DEFAULT_TEMPLATE_BRANCH_NAME
            )
            checkout_new_branch(repo, feature_branch_name)
            adder.apply_template_and_add(
                repo,
                cookiecutter_one_template,
                add_output_transaction,
                out_root=GENERATED_REPO_DIR,
                no_input=True,
                data=data,
                add_mode=AddMode.PROJECT,
                merged_branch_name=feature_merged_branch_name,
                template_branch_name=feature_template_branch_name,
            )
            repo.git.checkout("master")

    merger.merge_flexlate_branches(repo, "feature-1", delete=delete)
    merge_branch_into_current(repo, "feature-1")

    manual_commit_message = "Manually resolve conflicts"

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        # Prompt will be coming and may or may not have a temp repo location
        # Need to get reference to temp repo, parse message to do so
        pattern = re.compile(r"Successfully handled conflicts in ([\/\d\w]+)\?")
        match = pattern.search(prompt)

        if match:
            # This means we are operating on the temp_repo (template branch)
            tmp_dir = match.group(1)
            use_repo = Repo(tmp_dir)
        else:
            # This means we are operating on the main repo (merged branch)
            use_repo = repo

        stage_and_commit_all(use_repo, manual_commit_message)
        return True

    if commit_first:
        merge_branch_into_current(repo, "feature-2")
        stage_and_commit_all(repo, manual_commit_message)

    with patch.object(merger_mod, "confirm_user", _resolve_conflicts_then_type_yes):
        merger.merge_flexlate_branches(repo, "feature-2", delete=delete)

    if not commit_first:
        merge_branch_into_current(repo, "feature-2")
        stage_and_commit_all(repo, manual_commit_message)

    _assert_successful_merge(
        repo,
        get_flexlate_branch_name_for_feature_branch(
            "feature-2", DEFAULT_MERGED_BRANCH_NAME
        ),
        get_flexlate_branch_name_for_feature_branch(
            "feature-2", DEFAULT_TEMPLATE_BRANCH_NAME
        ),
        delete,
        main_branch_message=manual_commit_message,
    )


def _checkout_and_ensure_correct_commit(
    repo: Repo, branch_name: str, message: str = "Update flexlate templates"
):
    branch = repo.branches[branch_name]  # type: ignore
    branch.checkout()
    assert_main_commit_message_matches(message, repo.commit().message)


def _assert_successful_merge(
    repo: Repo,
    feature_merged_branch_name: str,
    feature_template_branch_name: str,
    delete: bool,
    feature_branch_message: str = "Update flexlate templates",
    main_branch_message: str = "Update flexlate templates",
):
    for branch_name in [
        "master",
        DEFAULT_MERGED_BRANCH_NAME,
        DEFAULT_TEMPLATE_BRANCH_NAME,
    ]:
        _checkout_and_ensure_correct_commit(
            repo, branch_name, message=main_branch_message
        )

    for branch_name in [feature_merged_branch_name, feature_template_branch_name]:
        if delete:
            with pytest.raises(IndexError):
                branch = repo.branches[branch_name]  # type: ignore
        else:
            _checkout_and_ensure_correct_commit(
                repo, branch_name, message=feature_branch_message
            )



from pathlib import Path
from typing import Optional, List, Sequence

from git import Repo

from flexlate.adder import Adder
from flexlate.add_mode import AddMode
from flexlate.bootstrapper import Bootstrapper
from flexlate.branch_update import get_flexlate_branch_name
from flexlate.checker import Checker, CheckResults, CheckResultsRenderable
from flexlate.config_manager import ConfigManager
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.finder.multi import MultiFinder
from flexlate.logger import log
from flexlate.merger import Merger
from flexlate.pusher import Pusher
from flexlate.remover import Remover
from flexlate.render.multi import MultiRenderer
from flexlate.styles import console
from flexlate.syncer import Syncer
from flexlate.template.base import Template
from flexlate.template_data import TemplateData
from flexlate.transactions.transaction import (
    FlexlateTransaction,
    TransactionType,
)
from flexlate.transactions.undoer import Undoer
from flexlate.update.main import Updater
from flexlate.user_config_manager import UserConfigManager

log.debug("Flexlate debug logging enabled")


class Flexlate:
    def __init__(
        self,
        quiet: bool = False,
        adder: Adder = Adder(),
        bootstrapper: Bootstrapper = Bootstrapper(),
        checker: Checker = Checker(),
        remover: Remover = Remover(),
        config_manager: ConfigManager = ConfigManager(),
        merger: Merger = Merger(),
        finder: MultiFinder = MultiFinder(),
        pusher: Pusher = Pusher(),
        renderer: MultiRenderer = MultiRenderer(),
        syncer: Syncer = Syncer(),
        undoer: Undoer = Undoer(),
        updater: Updater = Updater(),
        user_config_manager: UserConfigManager = UserConfigManager(),
    ):
        self.quiet = quiet
        # Let rich handle suppressing output
        console.quiet = quiet

        self.adder = adder
        self.bootstrapper = bootstrapper
        self.checker = checker
        self.remover = remover
        self.config_manager = config_manager
        self.merger = merger
        self.finder = finder
        self.pusher = pusher
        self.renderer = renderer
        self.syncer = syncer
        self.undoer = undoer
        self.updater = updater
        self.user_config_manager = user_config_manager

    def init_project(
        self,
        path: Path = Path("."),
        default_add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        user: bool = False,
        remote: str = "origin",
    ):
        repo = Repo(path)
        self.adder.init_project_and_add_to_branches(
            repo,
            default_add_mode=default_add_mode,
            user=user,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
            remote=remote,
            config_manager=self.config_manager,
        )

    def init_project_from(
        self,
        template_path: str,
        path: Path = Path("."),
        template_version: Optional[str] = None,
        data: Optional[TemplateData] = None,
        default_folder_name: str = "project",
        no_input: bool = False,
        default_add_mode: AddMode = AddMode.LOCAL,
        remote: str = "origin",
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    ) -> str:
        transaction = FlexlateTransaction(
            type=TransactionType.ADD_SOURCE_AND_OUTPUT,
            target=template_path,
            out_root=path,
        )
        template = self.finder.find(template_path, version=template_version)
        return self.adder.init_project_from_template_source_path(
            template,
            transaction,
            path=path,
            target_version=template_version,
            default_folder_name=default_folder_name,
            data=data,
            no_input=no_input,
            default_add_mode=default_add_mode,
            remote=remote,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
            config_manager=self.config_manager,
            updater=self.updater,
            renderer=self.renderer,
            syncer=self.syncer,
        )

    def add_template_source(
        self,
        path: str,
        name: Optional[str] = None,
        target_version: Optional[str] = None,
        template_root: Path = Path("."),
        add_mode: Optional[AddMode] = None,
    ):
        transaction = FlexlateTransaction(
            type=TransactionType.ADD_SOURCE, target=name or path, out_root=template_root
        )
        project_config = self.config_manager.load_project_config(template_root)
        add_mode = add_mode or project_config.default_add_mode
        template = self.finder.find(path, version=target_version)
        if name:
            template.name = name
        repo = Repo(project_config.path)
        self.adder.add_template_source(
            repo,
            template,
            transaction,
            target_version=target_version,
            out_root=template_root,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            add_mode=add_mode,
            remote=project_config.remote,
            config_manager=self.config_manager,
        )

    def remove_template_source(
        self,
        template_name: str,
        template_root: Path = Path("."),
    ):
        transaction = FlexlateTransaction(
            type=TransactionType.REMOVE_SOURCE,
            target=template_name,
            out_root=template_root,
        )
        project_config = self.config_manager.load_project_config(template_root)
        repo = Repo(project_config.path)
        self.remover.remove_template_source(
            repo,
            template_name,
            transaction,
            out_root=template_root,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            add_mode=project_config.default_add_mode,
            remote=project_config.remote,
            config_manager=self.config_manager,
        )

    def apply_template_and_add(
        self,
        name: str,
        data: Optional[TemplateData] = None,
        out_root: Path = Path("."),
        add_mode: Optional[AddMode] = None,
        no_input: bool = False,
    ):
        transaction = FlexlateTransaction(
            type=TransactionType.ADD_OUTPUT, target=name, out_root=out_root, data=data
        )
        project_config = self.config_manager.load_project_config(out_root)
        add_mode = add_mode or project_config.default_add_mode
        template = self.config_manager.get_template_by_name(
            name, project_root=project_config.path
        )
        repo = Repo(project_config.path)
        self.adder.apply_template_and_add(
            repo,
            template,
            transaction,
            data=data,
            out_root=out_root,
            add_mode=add_mode,
            no_input=no_input,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            remote=project_config.remote,
            config_manager=self.config_manager,
            renderer=self.renderer,
            updater=self.updater,
        )

    def remove_applied_template_and_output(
        self,
        template_name: str,
        out_root: Path = Path("."),
    ):
        transaction = FlexlateTransaction(
            type=TransactionType.REMOVE_OUTPUT, target=template_name, out_root=out_root
        )
        project_config = self.config_manager.load_project_config(out_root)
        repo = Repo(project_config.path)

        self.remover.remove_applied_template_and_output(
            repo,
            template_name,
            transaction,
            out_root=out_root,
            add_mode=project_config.default_add_mode,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            remote=project_config.remote,
            config_manager=self.config_manager,
            updater=self.updater,
            renderer=self.renderer,
        )

    def update(
        self,
        names: Optional[List[str]] = None,
        data: Optional[Sequence[TemplateData]] = None,
        no_input: bool = False,
        abort_on_conflict: bool = False,
        no_cleanup: bool = False,
        project_path: Path = Path("."),
    ):
        transaction = FlexlateTransaction(
            type=TransactionType.UPDATE,
            target=", ".join(names) if names is not None else None,
            data=data,
        )
        project_config = self.config_manager.load_project_config(project_path)
        templates: List[Template]
        if names:
            # User wants to update targeted template sources
            templates = [
                self.config_manager.get_template_by_name(
                    name, project_root=project_config.path
                )
                for name in names
            ]
        else:
            # User wants to update all templates
            templates = self.config_manager.get_all_templates(
                project_root=project_config.path
            )
        self.updater.update_passed_templates_to_target_versions(
            templates,
            project_root=project_config.path,
            finder=self.finder,
            config_manager=self.config_manager,
        )
        updates = self.updater.get_updates_for_templates(
            templates,
            data=data,
            config_manager=self.config_manager,
            project_root=project_config.path,
        )

        repo = Repo(project_config.path)
        self.updater.update(
            repo,
            updates,
            transaction,
            no_input=no_input,
            abort_on_conflict=abort_on_conflict,
            cleanup=not no_cleanup,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            remote=project_config.remote,
            renderer=self.renderer,
            config_manager=self.config_manager,
        )

    def undo(self, num_operations: int = 1, project_path: Path = Path(".")):
        project_config = self.config_manager.load_project_config(project_path)
        repo = Repo(project_config.path)
        self.undoer.undo_transactions(
            repo,
            num_transactions=num_operations,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
        )

    def sync(
        self,
        prompt: bool = False,
        project_path: Path = Path("."),
    ):
        project_config = self.config_manager.load_project_config(project_path)
        repo = Repo(project_config.path)
        transaction = FlexlateTransaction(
            type=TransactionType.SYNC,
        )
        self.syncer.sync_local_changes_to_flexlate_branches(
            repo,
            transaction,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            no_input=not prompt,
            remote=project_config.remote,
            updater=self.updater,
            renderer=self.renderer,
            config_manager=self.config_manager,
        )

    def merge_flexlate_branches(
        self,
        branch_name: Optional[str] = None,
        delete: bool = True,
        project_path: Path = Path("."),
    ):
        project_config = self.config_manager.load_project_config(project_path)
        repo = Repo(project_config.path)
        self.merger.merge_flexlate_branches(
            repo,
            branch_name,
            delete=delete,
            merged_branch_name=project_config.merged_branch_name,
            template_branch_name=project_config.template_branch_name,
        )

    def push_main_flexlate_branches(
        self,
        remote: Optional[str] = None,
        project_path: Path = Path("."),
    ):
        project_config = self.config_manager.load_project_config(project_path)
        use_remote = remote or project_config.remote
        repo = Repo(project_config.path)
        self.pusher.push_main_flexlate_branches(
            repo,
            use_remote,
            merged_branch_name=project_config.merged_branch_name,
            template_branch_name=project_config.template_branch_name,
        )

    def push_feature_flexlate_branches(
        self,
        feature_branch: Optional[str] = None,
        remote: str = "origin",
        project_path: Path = Path("."),
    ):
        project_config = self.config_manager.load_project_config(project_path)
        repo = Repo(project_config.path)
        self.pusher.push_feature_flexlate_branches(
            repo,
            feature_branch,
            remote,
            merged_branch_name=project_config.merged_branch_name,
            template_branch_name=project_config.template_branch_name,
        )

    def check(
        self, names: Optional[Sequence[str]] = None, project_path: Path = Path(".")
    ) -> CheckResults:
        check_results = self.checker.find_new_versions_for_template_sources(
            names,
            project_root=project_path,
            config_manager=self.config_manager,
            finder=self.finder,
        )
        console.print(CheckResultsRenderable(results=check_results.updates))
        return check_results

    def bootstrap_flexlate_init_from_existing_template(
        self,
        template_path: str,
        path: Path = Path("."),
        template_version: Optional[str] = None,
        data: Optional[TemplateData] = None,
        default_add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        remote: str = "origin",
        no_input: bool = False,
    ):
        transaction = FlexlateTransaction(
            type=TransactionType.BOOTSTRAP,
            target=template_path,
            out_root=path,
        )
        template = self.finder.find(template_path, version=template_version)
        repo = Repo(path)
        self.bootstrapper.bootstrap_flexlate_init_from_existing_template(
            repo,
            template,
            transaction,
            target_version=template_version,
            data=data,
            default_add_mode=default_add_mode,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
            remote=remote,
            no_input=no_input,
        )

    def update_template_source_target_version(
        self,
        name: str,
        target_version: Optional[str] = None,
        add_mode: Optional[AddMode] = None,
        project_path: Path = Path("."),
    ):
        project_config = self.config_manager.load_project_config(project_path)
        repo = Repo(project_config.path)
        add_mode = add_mode or project_config.default_add_mode
        transaction = FlexlateTransaction(
            type=TransactionType.UPDATE_TARGET_VERSION,
            target=name,
            data=dict(target_version=target_version),
        )

        self.user_config_manager.update_template_source_target_version(
            name,
            target_version,
            repo,
            transaction,
            project_path=project_path,
            add_mode=add_mode,
            merged_branch_name=get_flexlate_branch_name(
                repo, project_config.merged_branch_name
            ),
            base_merged_branch_name=project_config.merged_branch_name,
            template_branch_name=get_flexlate_branch_name(
                repo, project_config.template_branch_name
            ),
            base_template_branch_name=project_config.template_branch_name,
            remote=project_config.remote,
            config_manager=self.config_manager,
        )

    # TODO: list template sources, list applied templates



# A bootstrap script to set up input files for config tests
import os
from pathlib import Path

from flexlate.add_mode import AddMode
from flexlate.config import (
    FlexlateConfig,
    TemplateSource,
    AppliedTemplateConfig,
    FlexlateProjectConfig,
    ProjectConfig,
)
from flexlate.template.types import TemplateType
from tests.config import (
    COOKIECUTTER_ONE_DIR,
    CONFIGS_DIR,
    COOKIECUTTER_TWO_DIR,
    CONFIG_SUBDIR_2,
    PROJECT_DIR,
    PROJECT_CONFIGS_PROJECT_2_PATH,
    PROJECT_CONFIGS_ROOT_DIR,
    PROJECT_CONFIGS_PROJECT_1_PATH,
    PROJECT_CONFIGS_PROJECT_1_SUBDIR,
    PROJECT_CONFIGS_PROJECT_2_SUBDIR,
    GENERATED_FILES_DIR,
)


def create_config_1() -> FlexlateConfig:
    config = FlexlateConfig(
        template_sources=[
            TemplateSource(
                name="one",
                path=str(os.path.relpath(COOKIECUTTER_ONE_DIR, GENERATED_FILES_DIR)),
                type=TemplateType.COOKIECUTTER,
                render_relative_root_in_output=Path("{{ cookiecutter.a }}"),
                render_relative_root_in_template=Path("{{ cookiecutter.a }}"),
            ),
        ],
        applied_templates=[
            AppliedTemplateConfig(
                name="one",
                data=dict(a="b", c=""),
                version="d512c7e14e83cb4bc8d4e5ae06bb357e",
                add_mode=AddMode.LOCAL,
            ),
            AppliedTemplateConfig(
                name="one",
                data=dict(a="b", c=""),
                version="d512c7e14e83cb4bc8d4e5ae06bb357e",
                root=Path("subdir1"),
                add_mode=AddMode.PROJECT,
            ),
        ],
    )
    config.settings.custom_config_folder = CONFIGS_DIR
    return config


def create_config_2() -> FlexlateConfig:
    config = FlexlateConfig(
        template_sources=[
            TemplateSource(
                name="two",
                path=str(
                    os.path.relpath(
                        COOKIECUTTER_TWO_DIR, GENERATED_FILES_DIR / "subdir2"
                    )
                ),
                type=TemplateType.COOKIECUTTER,
                render_relative_root_in_output=Path("{{ cookiecutter.a }}"),
                render_relative_root_in_template=Path("{{ cookiecutter.a }}"),
            ),
        ],
        applied_templates=[
            AppliedTemplateConfig(
                name="two",
                data=dict(a="b", d="e"),
                version="some version",
                add_mode=AddMode.LOCAL,
            ),
            AppliedTemplateConfig(
                name="one",
                data=dict(a="b", c="something"),
                version="d512c7e14e83cb4bc8d4e5ae06bb357e",
                root=Path("subdir2_2"),
                add_mode=AddMode.LOCAL,
            ),
        ],
    )
    config.settings.custom_config_folder = CONFIG_SUBDIR_2
    return config


def gen_configs():
    config_1 = create_config_1()
    config_1.save()
    config_2 = create_config_2()
    config_2.save()


def create_project_config_root() -> FlexlateProjectConfig:
    config = FlexlateProjectConfig(
        projects=[
            ProjectConfig(
                path=PROJECT_CONFIGS_PROJECT_2_PATH.relative_to(
                    PROJECT_CONFIGS_ROOT_DIR
                )
            )
        ]
    )
    config.settings.custom_config_folder = PROJECT_CONFIGS_ROOT_DIR
    return config


def create_project_config_1() -> FlexlateProjectConfig:
    config = FlexlateProjectConfig(projects=[ProjectConfig(path=Path("."))])
    config.settings.custom_config_folder = PROJECT_CONFIGS_PROJECT_1_PATH
    return config


def gen_project_configs():
    if not PROJECT_CONFIGS_PROJECT_1_SUBDIR.exists():
        PROJECT_CONFIGS_PROJECT_1_SUBDIR.mkdir(parents=True)
    if not PROJECT_CONFIGS_PROJECT_2_SUBDIR.exists():
        PROJECT_CONFIGS_PROJECT_2_SUBDIR.mkdir(parents=True)
    config_root = create_project_config_root()
    config_root.save()
    config_1 = create_project_config_1()
    config_1.save()


if __name__ == "__main__":
    gen_configs()
    gen_project_configs()



from git import Repo

from flexlate import exc
from flexlate.config_manager import ConfigManager
from flexlate.constants import DEFAULT_TEMPLATE_BRANCH_NAME, DEFAULT_MERGED_BRANCH_NAME
from flexlate.render.multi import MultiRenderer
from flexlate.styles import print_styled, INFO_STYLE, SUCCESS_STYLE
from flexlate.transactions.transaction import FlexlateTransaction
from flexlate.update.main import Updater


class Syncer:
    def sync_local_changes_to_flexlate_branches(
        self,
        repo: Repo,
        transaction: FlexlateTransaction,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        no_input: bool = False,
        remote: str = "origin",
        updater: Updater = Updater(),
        renderer: MultiRenderer = MultiRenderer(),
        config_manager: ConfigManager = ConfigManager(),
    ):
        print_styled("Syncing local changes to flexlate branches", INFO_STYLE)
        try:
            updater.update(
                repo,
                [],
                transaction,
                merged_branch_name=merged_branch_name,
                base_merged_branch_name=base_merged_branch_name,
                template_branch_name=template_branch_name,
                base_template_branch_name=base_template_branch_name,
                no_input=no_input,
                full_rerender=True,
                remote=remote,
                renderer=renderer,
                config_manager=config_manager,
            )
        except exc.TriedToCommitButNoChangesException as e:
            raise exc.UnnecessarySyncException("Everything is up to date") from e
        print_styled(
            "Successfully synced local changes to flexlate branches", SUCCESS_STYLE
        )



from pathlib import Path
from typing import Optional

from flexlate.template.base import Template
from flexlate.template.types import TemplateType
from flexlate.template_config.cookiecutter import CookiecutterConfig


class CookiecutterTemplate(Template):
    _type = TemplateType.COOKIECUTTER

    def __init__(
        self,
        config: CookiecutterConfig,
        path: Path,
        render_relative_root: Path,
        name: Optional[str] = None,
        version: Optional[str] = None,
        target_version: Optional[str] = None,
        git_url: Optional[str] = None,
        template_source_path: Optional[str] = None,
    ):
        super().__init__(
            config,
            path,
            name=name,
            version=version,
            target_version=target_version,
            git_url=git_url,
            template_source_path=template_source_path,
            # Output relative root is always the same as template relative root for Cookiecutter
            render_relative_root_in_output=render_relative_root,
            render_relative_root_in_template=render_relative_root,
        )



import shlex
import subprocess
from typing import Dict, Optional, Union, List

from flexlate.template_data import TemplateData


def run(
    command: str, input_data: Optional[Dict[str, str]] = None
) -> subprocess.CompletedProcess:
    split_command = shlex.split(command)
    input = get_text_input_from_data(input_data)
    return subprocess.run(split_command, check=True, input=input, encoding="utf8")


def get_text_input_from_data(
    input_data: Optional[Union[TemplateData, List[TemplateData]]]
) -> Optional[str]:
    if input_data is None:
        return None
    if isinstance(input_data, list):
        return "\n".join(_template_data_to_text_input(data) for data in input_data)
    return _template_data_to_text_input(input_data)


def _template_data_to_text_input(input_data: TemplateData) -> str:
    return "\n".join(str(value) for value in input_data.values())



import contextlib
from typing import Sequence

from git import Repo

from flexlate.pusher import Pusher
from tests.config import GENERATED_FILES_DIR
from tests.dirutils import assert_dir_trees_are_equal
from tests.fixtures.templated_repo import *
from tests.gitutils import add_remote, add_local_remote


def test_push_feature_flexlate_branches(
    repo_with_template_branch_from_cookiecutter_one_and_feature_flexlate_branches: Repo,
):
    repo = repo_with_template_branch_from_cookiecutter_one_and_feature_flexlate_branches
    feature_merged_branch_name = get_flexlate_branch_name(
        repo, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name(
        repo, DEFAULT_TEMPLATE_BRANCH_NAME
    )
    pusher = Pusher()
    with add_local_remote_and_check_branches_on_exit(
        repo, [feature_merged_branch_name, feature_template_branch_name]
    ):
        pusher.push_feature_flexlate_branches(repo)


def test_push_main_flexlate_branches(
    repo_with_template_branch_from_cookiecutter_one: Repo,
):
    repo = repo_with_template_branch_from_cookiecutter_one
    pusher = Pusher()
    with add_local_remote_and_check_branches_on_exit(
        repo, [DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME]
    ):
        pusher.push_main_flexlate_branches(repo)


@contextlib.contextmanager
def add_local_remote_and_check_branches_on_exit(
    repo: Repo, branch_names: Sequence[str]
):
    remote_path = GENERATED_FILES_DIR / "remote"
    remote_repo = add_local_remote(repo, remote_path=remote_path)

    yield

    for branch_name in branch_names:
        branch = repo.branches[branch_name]  # type: ignore
        remote_branch = remote_repo.branches[branch_name]  # type: ignore
        branch.checkout()
        remote_branch.checkout()
        assert_dir_trees_are_equal(GENERATED_REPO_DIR, remote_path)
